import numpy as np
import numba as nb

from sigpy import backend, config


__all__ = ['array_to_blocks', 'blocks_to_array']


def array_to_blocks(input, blk_shape, blk_strides):
    """Extract blocks from an array in a sliding window manner.

    Args:
        input (array): input array. input.ndim must be <= 4.
        blk_shape (tuple): block shape. Must have same length as input.ndim.
        blk_strides (tuple): block strides.
            Must have same length as input.ndim.

    Returns:
        array: array of shape num_blks + blk_shape, where
        num_blks = (input.shape - blk_shape + blk_strides) // blk_strides

    Example:

        >>> input = np.array([0, 1, 2, 3, 4, 5])
        >>> print(array_to_blocks(input, [2], [2]))
        [[0, 1],
         [2, 3],
         [4, 5]]

    """
    ndim = input.ndim

    if ndim != len(blk_shape) or ndim != len(blk_strides):
        raise ValueError('Input must have same dimensions as blocks.')

    num_blks = [(i - b + s) // s for i, b,
                s in zip(input.shape, blk_shape, blk_strides)]
    device = backend.get_device(input)
    xp = device.xp
    with device:
        output = xp.zeros(num_blks + blk_shape, dtype=input.dtype)

        if ndim == 1:
            if device == backend.cpu_device:
                _array_to_blocks1(output, input,
                                  blk_shape[-1],
                                  blk_strides[-1],
                                  num_blks[-1])
            else:  # pragma: no cover
                _array_to_blocks1_cuda(input,
                                       blk_shape[-1],
                                       blk_strides[-1],
                                       num_blks[-1],
                                       output,
                                       size=num_blks[-1] * blk_shape[-1])
        elif ndim == 2:
            if device == backend.cpu_device:
                _array_to_blocks2(output, input,
                                  blk_shape[-1], blk_shape[-2],
                                  blk_strides[-1], blk_strides[-2],
                                  num_blks[-1], num_blks[-2])
            else:  # pragma: no cover
                _array_to_blocks2_cuda(input,
                                       blk_shape[-1], blk_shape[-2],
                                       blk_strides[-1], blk_strides[-2],
                                       num_blks[-1], num_blks[-2],
                                       output,
                                       size=num_blks[-1] * num_blks[-2] *
                                       blk_shape[-1] * blk_shape[-2])
        elif ndim == 3:
            if device == backend.cpu_device:
                _array_to_blocks3(output,
                                  input,
                                  blk_shape[-1],
                                  blk_shape[-2],
                                  blk_shape[-3],
                                  blk_strides[-1],
                                  blk_strides[-2],
                                  blk_strides[-3],
                                  num_blks[-1],
                                  num_blks[-2],
                                  num_blks[-3])
            else:  # pragma: no cover
                _array_to_blocks3_cuda(input,
                                       blk_shape[-1], blk_shape[-2],
                                       blk_shape[-3],
                                       blk_strides[-1], blk_strides[-2],
                                       blk_strides[-3],
                                       num_blks[-1], num_blks[-2],
                                       num_blks[-3],
                                       output,
                                       size=num_blks[-1] * num_blks[-2] *
                                       num_blks[-3] *
                                       blk_shape[-1] * blk_shape[-2] *
                                       blk_shape[-3])
        elif ndim == 4:
            if device == backend.cpu_device:
                _array_to_blocks4(output,
                                  input,
                                  blk_shape[-1],
                                  blk_shape[-2],
                                  blk_shape[-3],
                                  blk_shape[-4],
                                  blk_strides[-1],
                                  blk_strides[-2],
                                  blk_strides[-3],
                                  blk_strides[-4],
                                  num_blks[-1],
                                  num_blks[-2],
                                  num_blks[-3],
                                  num_blks[-4])
            else:  # pragma: no cover
                _array_to_blocks4_cuda(input,
                                       blk_shape[-1], blk_shape[-2],
                                       blk_shape[-3], blk_shape[-4],
                                       blk_strides[-1], blk_strides[-2],
                                       blk_strides[-3], blk_strides[-4],
                                       num_blks[-1], num_blks[-2],
                                       num_blks[-3], num_blks[-4],
                                       output,
                                       size=num_blks[-1] * num_blks[-2] *
                                       num_blks[-3] * num_blks[-4] *
                                       blk_shape[-1] * blk_shape[-2] *
                                       blk_shape[-3] * blk_shape[-4])
        else:
            raise ValueError('Only support input.ndim <= 4'
                             ', got {}'.format(ndim))

        return output


def blocks_to_array(input, oshape, blk_shape, blk_strides):
    """Accumulate blocks into an array in a sliding window manner.

    Args:
        input (array): input array of shape num_blks + blk_shape
        oshape (tuple): output shape.
        blk_shape (tuple): block shape. Must have same length as oshape.
        blk_strides (tuple): block strides. Must have same length as oshape.

    Returns:
        array: array of shape oshape.

    """
    ndim = len(blk_shape)

    if 2 * ndim != input.ndim or ndim != len(blk_strides):
        raise ValueError('Input must have same dimensions as blocks.')

    num_blks = input.shape[:ndim]
    device = backend.get_device(input)
    xp = device.xp
    with device:
        output = xp.zeros(oshape, dtype=input.dtype)

        if ndim == 1:
            if device == backend.cpu_device:
                _blocks_to_array1(output, input,
                                  blk_shape[-1],
                                  blk_strides[-1],
                                  num_blks[-1])
            else:  # pragma: no cover
                if np.issubdtype(input.dtype, np.floating):
                    _blocks_to_array1_cuda(input,
                                           blk_shape[-1],
                                           blk_strides[-1],
                                           num_blks[-1],
                                           output,
                                           size=num_blks[-1] * blk_shape[-1])
                else:
                    _blocks_to_array1_cuda_complex(input,
                                                   blk_shape[-1],
                                                   blk_strides[-1],
                                                   num_blks[-1],
                                                   output,
                                                   size=num_blks[-1] *
                                                   blk_shape[-1])
        elif ndim == 2:
            if device == backend.cpu_device:
                _blocks_to_array2(output, input,
                                  blk_shape[-1], blk_shape[-2],
                                  blk_strides[-1], blk_strides[-2],
                                  num_blks[-1], num_blks[-2])
            else:  # pragma: no cover
                if np.issubdtype(input.dtype, np.floating):
                    _blocks_to_array2_cuda(input,
                                           blk_shape[-1], blk_shape[-2],
                                           blk_strides[-1], blk_strides[-2],
                                           num_blks[-1], num_blks[-2],
                                           output,
                                           size=num_blks[-1] * num_blks[-2] *
                                           blk_shape[-1] * blk_shape[-2])
                else:  # pragma: no cover
                    _blocks_to_array2_cuda_complex(
                        input,
                        blk_shape[-1], blk_shape[-2],
                        blk_strides[-1], blk_strides[-2],
                        num_blks[-1], num_blks[-2],
                        output,
                        size=num_blks[-1] * num_blks[-2] *
                        blk_shape[-1] * blk_shape[-2])
        elif ndim == 3:
            if device == backend.cpu_device:
                _blocks_to_array3(output,
                                  input,
                                  blk_shape[-1],
                                  blk_shape[-2],
                                  blk_shape[-3],
                                  blk_strides[-1],
                                  blk_strides[-2],
                                  blk_strides[-3],
                                  num_blks[-1],
                                  num_blks[-2],
                                  num_blks[-3])
            else:  # pragma: no cover
                if np.issubdtype(input.dtype, np.floating):
                    _blocks_to_array3_cuda(
                        input,
                        blk_shape[-1], blk_shape[-2], blk_shape[-3],
                        blk_strides[-1], blk_strides[-2], blk_strides[-3],
                        num_blks[-1], num_blks[-2], num_blks[-3],
                        output,
                        size=num_blks[-1] * num_blks[-2] *
                        num_blks[-3] * blk_shape[-1] * blk_shape[-2] *
                        blk_shape[-3])
                else:
                    _blocks_to_array3_cuda_complex(
                        input,
                        blk_shape[-1], blk_shape[-2], blk_shape[-3],
                        blk_strides[-1], blk_strides[-2],
                        blk_strides[-3],
                        num_blks[-1], num_blks[-2], num_blks[-3],
                        output,
                        size=num_blks[-1] * num_blks[-2] * num_blks[-3] *
                        blk_shape[-1] * blk_shape[-2] * blk_shape[-3])
        elif ndim == 4:
            if device == backend.cpu_device:
                _blocks_to_array4(output,
                                  input,
                                  blk_shape[-1],
                                  blk_shape[-2],
                                  blk_shape[-3],
                                  blk_shape[-4],
                                  blk_strides[-1],
                                  blk_strides[-2],
                                  blk_strides[-3],
                                  blk_strides[-4],
                                  num_blks[-1],
                                  num_blks[-2],
                                  num_blks[-3],
                                  num_blks[-4])
            else:  # pragma: no cover
                if np.issubdtype(input.dtype, np.floating):
                    _blocks_to_array4_cuda(
                        input,
                        blk_shape[-1], blk_shape[-2],
                        blk_shape[-3], blk_shape[-4],
                        blk_strides[-1], blk_strides[-2],
                        blk_strides[-3], blk_strides[-4],
                        num_blks[-1], num_blks[-2],
                        num_blks[-3], num_blks[-4],
                        output,
                        size=num_blks[-1] * num_blks[-2] *
                        num_blks[-3] * num_blks[-4] *
                        blk_shape[-1] * blk_shape[-2] *
                        blk_shape[-3] * blk_shape[-4])
                else:
                    _blocks_to_array4_cuda_complex(
                        input,
                        blk_shape[-1], blk_shape[-2],
                        blk_shape[-3], blk_shape[-4],
                        blk_strides[-1], blk_strides[-2],
                        blk_strides[-3], blk_strides[-4],
                        num_blks[-1], num_blks[-2],
                        num_blks[-3], num_blks[-4],
                        output,
                        size=num_blks[-1] * num_blks[-2] *
                        num_blks[-3] * num_blks[-4] *
                        blk_shape[-1] * blk_shape[-2] *
                        blk_shape[-3] * blk_shape[-4])
        else:
            raise ValueError('Only support input.ndim <= 4'
                             ', got {}'.format(ndim))

        return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks1(output, input, Bx, Sx, Nx):
    for nx in range(Nx):
        for bx in range(Bx):
            ix = nx * Sx + bx
            if ix < input.shape[-1]:
                output[nx, bx] = input[ix]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks2(output, input, Bx, By, Sx, Sy, Nx, Ny):
    for ny in range(Ny):
        for nx in range(Nx):
            for by in range(By):
                for bx in range(Bx):
                    iy = ny * Sy + by
                    ix = nx * Sx + bx
                    if ix < input.shape[-1] and iy < input.shape[-2]:
                        output[ny, nx, by, bx] = input[iy, ix]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks3(output, input, Bx, By, Bz, Sx, Sy, Sz, Nx, Ny, Nz):
    for nz in range(Nz):
        for ny in range(Ny):
            for nx in range(Nx):
                for bz in range(Bz):
                    for by in range(By):
                        for bx in range(Bx):
                            iz = nz * Sz + bz
                            iy = ny * Sy + by
                            ix = nx * Sx + bx
                            if (ix < input.shape[-1] and
                                iy < input.shape[-2] and
                                iz < input.shape[-3]):
                                output[nz, ny, nx, bz, by,
                                       bx] = input[iz, iy, ix]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks4(output, input, Bx, By, Bz, Bt, Sx, Sy, Sz, St,
                      Nx, Ny, Nz, Nt):
    for nt in range(Nt):
        for nz in range(Nz):
            for ny in range(Ny):
                for nx in range(Nx):
                    for bt in range(Bt):
                        for bz in range(Bz):
                            for by in range(By):
                                for bx in range(Bx):
                                    it = nt * St + bt
                                    iz = nz * Sz + bz
                                    iy = ny * Sy + by
                                    ix = nx * Sx + bx
                                    if (ix < input.shape[-1] and
                                        iy < input.shape[-2] and
                                        iz < input.shape[-3] and
                                        it < input.shape[-4]):
                                        output[nt, nz, ny, nx,
                                               bt, bz, by,
                                               bx] = input[it, iz, iy, ix]


@nb.jit(nopython=True, cache=True)
def _blocks_to_array1(output, input, Bx, Sx, Nx):  # pragma: no cover
    for nx in range(Nx):
        for bx in range(Bx):
            ix = nx * Sx + bx
            if ix < output.shape[-1]:
                output[ix] += input[nx, bx]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array2(output, input, Bx, By, Sx, Sy, Nx, Ny):
    for ny in range(Ny):
        for nx in range(Nx):
            for by in range(By):
                for bx in range(Bx):
                    iy = ny * Sy + by
                    ix = nx * Sx + bx
                    if ix < output.shape[-1] and iy < output.shape[-2]:
                        output[iy, ix] += input[ny, nx, by, bx]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array3(output, input, Bx, By, Bz, Sx, Sy, Sz, Nx, Ny, Nz):
    for nz in range(Nz):
        for ny in range(Ny):
            for nx in range(Nx):
                for bz in range(Bz):
                    for by in range(By):
                        for bx in range(Bx):
                            iz = nz * Sz + bz
                            iy = ny * Sy + by
                            ix = nx * Sx + bx
                            if (ix < output.shape[-1]
                                and iy < output.shape[-2]
                                and iz < output.shape[-3]):
                                output[iz, iy, ix] += input[nz,
                                                            ny, nx, bz, by, bx]


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array4(output, input, Bx, By, Bz, Bt,
                      Sx, Sy, Sz, St, Nx, Ny, Nz, Nt):
    for nt in range(Nt):
        for nz in range(Nz):
            for ny in range(Ny):
                for nx in range(Nx):
                    for bt in range(Bt):
                        for bz in range(Bz):
                            for by in range(By):
                                for bx in range(Bx):
                                    it = nt * St + bt
                                    iz = nz * Sz + bz
                                    iy = ny * Sy + by
                                    ix = nx * Sx + bx
                                    if (ix < output.shape[-1]
                                        and iy < output.shape[-2]
                                        and iz < output.shape[-3]
                                        and it < output.shape[-4]):
                                        output[it,
                                               iz,
                                               iy,
                                               ix] += input[nt,
                                                            nz,
                                                            ny,
                                                            nx,
                                                            bt,
                                                            bz,
                                                            by,
                                                            bx]


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _array_to_blocks1_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 Sx, int32 Nx',
        'raw T output',
        """
        const int ndim = input.ndim;

        int nx = i / Bx;
        i -= nx * Bx;
        int bx = i;

        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1]) {
            int input_idx[] = {ix};
            int output_idx[] = {nx, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name='_array_to_blocks1_cuda')

    _array_to_blocks2_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, '
        'int32 Sx, int32 Sy, int32 Nx, int32 Ny',
        'raw T output',
        """
        const int ndim = input.ndim;

        int ny = i / Bx / By / Nx;
        i -= ny * Bx * By * Nx;
        int nx = i / Bx / By;
        i -= nx * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1] && iy < input.shape()[ndim - 2]) {
            int input_idx[] = {iy, ix};
            int output_idx[] = {ny, nx, by, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name='_array_to_blocks2_cuda')

    _array_to_blocks3_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, '
        'int32 Sx, int32 Sy, int32 Sz, int32 Nx, int32 Ny, int32 Nz',
        'raw T output',
        """
        const int ndim = input.ndim;

        int nz = i / Bx / By / Bz / Nx / Ny;
        i -= nz * Bx * By * Bz * Nx * Ny;
        int ny = i / Bx / By / Bz / Nx;
        i -= ny * Bx * By * Bz * Nx;
        int nx = i / Bx / By / Bz;
        i -= nx * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1] && iy < input.shape()[ndim - 2]
            && iz < input.shape()[ndim - 3]) {
            int input_idx[] = {iz, iy, ix};
            int output_idx[] = {nz, ny, nx, bz, by, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name='_array_to_blocks3_cuda')

    _array_to_blocks4_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, int32 Bt, '
        'int32 Sx, int32 Sy, int32 Sz, int32 St, '
        'int32 Nx, int32 Ny, int32 Nz, int32 Nt',
        'raw T output',
        """
        const int ndim = input.ndim;

        int nt = i / Bx / By / Bz / Bt / Nx / Ny / Nz;
        i -= nt * Bx * By * Bz * Bt * Nx * Ny * Nz;
        int nz = i / Bx / By / Bz / Bt / Nx / Ny;
        i -= nz * Bx * By * Bz * Bt * Nx * Ny;
        int ny = i / Bx / By / Bz / Bt / Nx;
        i -= ny * Bx * By * Bz * Bt * Nx;
        int nx = i / Bx / By / Bz / Bt;
        i -= nx * Bx * By * Bz * Bt;
        int bt = i / Bx / By / Bz;
        i -= bt * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int it = nt * St + bt;
        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1] && iy < input.shape()[ndim - 2]
            && iz < input.shape()[ndim - 3] && it < input.shape()[ndim - 4]) {
            int input_idx[] = {it, iz, iy, ix};
            int output_idx[] = {nt, nz, ny, nx, bt, bz, by, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name='_array_to_blocks4_cuda')

    _blocks_to_array1_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 Sx, int32 Nx',
        'raw T output',
        """
        const int ndim = output.ndim;

        int nx = i / Bx;
        i -= nx * Bx;
        int bx = i;

        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1]) {
            int input_idx[] = {nx, bx};
            int output_idx[] = {ix};
            atomicAdd(&output[output_idx], input[input_idx]);
        }
        """,
        name='_blocks_to_array1_cuda')

    _blocks_to_array1_cuda_complex = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 Sx, int32 Nx',
        'raw T output',
        """
        const int ndim = output.ndim;

        int nx = i / Bx;
        i -= nx * Bx;
        int bx = i;

        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1]) {
            int input_idx[] = {nx, bx};
            int output_idx[] = {ix};
            atomicAdd(reinterpret_cast<T::value_type*>(&(output[output_idx])),
                input[input_idx].real());
            atomicAdd(
                reinterpret_cast<T::value_type*>(&(output[output_idx])) + 1,
                input[input_idx].imag());
        }
        """,
        name='_blocks_to_array1_cuda_complex')

    _blocks_to_array2_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Sx, int32 Sy, '
        'int32 Nx, int32 Ny',
        'raw T output',
        """
        const int ndim = output.ndim;

        int ny = i / Bx / By / Nx;
        i -= ny * Bx * By * Nx;
        int nx = i / Bx / By;
        i -= nx * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1] && iy < output.shape()[ndim - 2]) {
            int input_idx[] = {ny, nx, by, bx};
            int output_idx[] = {iy, ix};
            atomicAdd(&output[output_idx], input[input_idx]);
        }
        """,
        name='_blocks_to_array2_cuda')

    _blocks_to_array2_cuda_complex = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Sx, int32 Sy, '
        'int32 Nx, int32 Ny',
        'raw T output',
        """
        const int ndim = output.ndim;

        int ny = i / Bx / By / Nx;
        i -= ny * Bx * By * Nx;
        int nx = i / Bx / By;
        i -= nx * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1] && iy < output.shape()[ndim - 2]) {
            int input_idx[] = {ny, nx, by, bx};
            int output_idx[] = {iy, ix};
            atomicAdd(reinterpret_cast<T::value_type*>(&(output[output_idx])),
                input[input_idx].real());
            atomicAdd(
                reinterpret_cast<T::value_type*>(&(output[output_idx])) + 1,
                input[input_idx].imag());
        }
        """,
        name='_blocks_to_array2_cuda_complex')

    _blocks_to_array3_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, '
        'int32 Sx, int32 Sy, int32 Sz, int32 Nx, int32 Ny, int32 Nz',
        'raw T output',
        """
        const int ndim = output.ndim;

        int nz = i / Bx / By / Bz / Nx / Ny;
        i -= nz * Bx * By * Bz * Nx * Ny;
        int ny = i / Bx / By / Bz / Nx;
        i -= ny * Bx * By * Bz * Nx;
        int nx = i / Bx / By / Bz;
        i -= nx * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1] &&
            iy < output.shape()[ndim - 2] &&
            iz < output.shape()[ndim - 3]) {
            int input_idx[] = {nz, ny, nx, bz, by, bx};
            int output_idx[] = {iz, iy, ix};
            atomicAdd(&output[output_idx], input[input_idx]);
        }
        """,
        name='_blocks_to_array3_cuda')

    _blocks_to_array3_cuda_complex = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, '
        'int32 Sx, int32 Sy, int32 Sz, int32 Nx, int32 Ny, int32 Nz',
        'raw T output',
        """
        const int ndim = output.ndim;

        int nz = i / Bx / By / Bz / Nx / Ny;
        i -= nz * Bx * By * Bz * Nx * Ny;
        int ny = i / Bx / By / Bz / Nx;
        i -= ny * Bx * By * Bz * Nx;
        int nx = i / Bx / By / Bz;
        i -= nx * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < output.shape()[ndim - 1] &&
            iy < output.shape()[ndim - 2] &&
            iz < output.shape()[ndim - 3]) {
            int input_idx[] = {nz, ny, nx, bz, by, bx};
            int output_idx[] = {iz, iy, ix};
            atomicAdd(reinterpret_cast<T::value_type*>(&(output[output_idx])),
                input[input_idx].real());
            atomicAdd(reinterpret_cast<T::value_type*>(
                &(output[output_idx])) + 1, input[input_idx].imag());
        }
        """,
        name='_blocks_to_array3_cuda_complex')

    _blocks_to_array4_cuda = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, int32 Bt, '
        'int32 Sx, int32 Sy, int32 Sz, int32 St, '
        'int32 Nx, int32 Ny, int32 Nz, int32 Nt',
        'raw T output',
        """
        const int ndim = output.ndim;
        int nt = i / Bx / By / Bz / Bt / Nx / Ny / Nz;
        i -= nt * Bx * By * Bz * Bt * Nx * Ny * Nz;
        int nz = i / Bx / By / Bz / Bt / Nx / Ny;
        i -= nz * Bx * By * Bz * Bt * Nx * Ny;
        int ny = i / Bx / By / Bz / Bt / Nx;
        i -= ny * Bx * By * Bz * Bt * Nx;
        int nx = i / Bx / By / Bz / Bt;
        i -= nx * Bx * By * Bz * Bt;
        int bt = i / Bx / By / Bz;
        i -= bt * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int it = nt * St + bt;
        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;

        if (ix < output.shape()[ndim - 1] &&
            iy < output.shape()[ndim - 2] &&
            iz < output.shape()[ndim - 3] &&
            it < output.shape()[ndim - 4]) {
            int input_idx[] = {nt, nz, ny, nx, bt, bz, by, bx};
            int output_idx[] = {it, iz, iy, ix};
            atomicAdd(&output[output_idx], input[input_idx]);
        }
        """,
        name='_blocks_to_array4_cuda')

    _blocks_to_array4_cuda_complex = cp.ElementwiseKernel(
        'raw T input, int32 Bx, int32 By, int32 Bz, int32 Bt, '
        'int32 Sx, int32 Sy, int32 Sz, int32 St, '
        'int32 Nx, int32 Ny, int32 Nz, int32 Nt',
        'raw T output',
        """
        const int ndim = output.ndim;
        int nt = i / Bx / By / Bz / Bt / Nx / Ny / Nz;
        i -= nt * Bx * By * Bz * Bt * Nx * Ny * Nz;
        int nz = i / Bx / By / Bz / Bt / Nx / Ny;
        i -= nz * Bx * By * Bz * Bt * Nx * Ny;
        int ny = i / Bx / By / Bz / Bt / Nx;
        i -= ny * Bx * By * Bz * Bt * Nx;
        int nx = i / Bx / By / Bz / Bt;
        i -= nx * Bx * By * Bz * Bt;
        int bt = i / Bx / By / Bz;
        i -= bt * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int it = nt * St + bt;
        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;

        if (ix < output.shape()[ndim - 1] &&
            iy < output.shape()[ndim - 2] &&
            iz < output.shape()[ndim - 3] &&
            it < output.shape()[ndim - 4]) {
            int input_idx[] = {nt, nz, ny, nx, bt, bz, by, bx};
            int output_idx[] = {it, iz, iy, ix};
            atomicAdd(reinterpret_cast<T::value_type*>(&(output[output_idx])),
                input[input_idx].real());
            atomicAdd(reinterpret_cast<T::value_type*>(
                &(output[output_idx])) + 1, input[input_idx].imag());
        }
        """,
        name='_blocks_to_array4_cuda_complex')



import unittest
import numpy as np
import sigpy as sp
import numpy.testing as npt

from sigpy.mri import linop, precond

if __name__ == '__main__':
    unittest.main()


class TestPrecond(unittest.TestCase):

    def test_kspace_precond_cart(self):
        nc = 4
        n = 10
        shape = (nc, n)
        mps = sp.randn(shape, dtype=np.complex)
        mps /= np.linalg.norm(mps, axis=0, keepdims=True)
        weights = sp.randn([n]) >= 0

        A = sp.linop.Multiply(shape, weights**0.5) * linop.Sense(mps)

        AAH = np.zeros((nc, n, nc, n), np.complex)
        for d in range(nc):
            for j in range(n):
                x = np.zeros((nc, n), np.complex)
                x[d, j] = 1.0
                AAHx = A(A.H(x))

                for c in range(nc):
                    for i in range(n):
                        AAH[c, i, d, j] = AAHx[c, i]

        p_expected = np.ones((nc, n), np.complex)
        for c in range(nc):
            for i in range(n):
                if weights[i]:
                    p_expected_inv_ic = 0
                    for d in range(nc):
                        for j in range(n):
                            p_expected_inv_ic += abs(
                                AAH[c, i, d, j])**2 / abs(AAH[c, i, c, i])

                    p_expected[c, i] = 1 / p_expected_inv_ic

        p = precond.kspace_precond(mps, weights=weights)
        npt.assert_allclose(p[:, weights == 1], p_expected[:, weights == 1])

    def test_kspace_precond_noncart(self):
        n = 10
        nc = 3
        shape = [nc, n]
        mps = sp.randn(shape, dtype=np.complex)
        mps /= np.linalg.norm(mps, axis=0, keepdims=True)
        coord = sp.randn([n, 1], dtype=np.float)

        A = linop.Sense(mps, coord=coord)

        AAH = np.zeros((nc, n, nc, n), np.complex)
        for d in range(nc):
            for j in range(n):
                x = np.zeros(shape, np.complex)
                x[d, j] = 1.0
                AAHx = A(A.H(x))
                for c in range(nc):
                    for i in range(n):
                        AAH[c, i, d, j] = AAHx[c, i]

        p_expected = np.zeros([nc, n], np.complex)
        for c in range(nc):
            for i in range(n):
                p_expected_inv_ic = 0
                for d in range(nc):
                    for j in range(n):
                        p_expected_inv_ic += abs(AAH[c, i, d, j]
                                                 )**2 / abs(AAH[c, i, c, i])

                p_expected[c, i] = 1 / p_expected_inv_ic

        p = precond.kspace_precond(mps, coord=coord)
        npt.assert_allclose(p, p_expected, atol=1e-2, rtol=1e-2)

    def test_kspace_precond_simple_cart(self):
        # Check identity
        mps_shape = [1, 1]
        mps = np.ones(mps_shape, dtype=np.complex)
        p = precond.kspace_precond(mps)
        npt.assert_allclose(p, np.ones(mps_shape))

        # Check scaling
        mps_shape = [1, 3]
        mps = np.ones(mps_shape, dtype=np.complex)
        p = precond.kspace_precond(mps)
        npt.assert_allclose(p, np.ones(mps_shape))

        # Check 2d
        mps_shape = [1, 3, 3]
        mps = np.ones(mps_shape, dtype=np.complex)
        p = precond.kspace_precond(mps)
        npt.assert_allclose(p, np.ones(mps_shape))

        # Check weights
        mps_shape = [1, 3]
        mps = np.ones(mps_shape, dtype=np.complex)
        weights = np.array([1, 0, 1], dtype=np.complex)
        p = precond.kspace_precond(mps, weights=weights)
        npt.assert_allclose(p, [[1, 1, 1]])

    def test_kspace_precond_simple_noncart(self):
        # Check identity
        mps_shape = [1, 1]

        mps = np.ones(mps_shape, dtype=np.complex)
        coord = np.array([[0.0]])
        p = precond.kspace_precond(mps, coord=coord)
        npt.assert_allclose(p, [[1.0]], atol=1, rtol=1e-1)

        mps_shape = [1, 3]

        mps = np.ones(mps_shape, dtype=np.complex)
        coord = np.array([[0.0], [-1], [1]])
        p = precond.kspace_precond(mps, coord=coord)
        npt.assert_allclose(p, [[1.0, 1.0, 1.0]], atol=1, rtol=1e-1)

    def test_circulant_precond_cart(self):
        nc = 4
        n = 10
        shape = (nc, n)
        mps = sp.randn(shape, dtype=np.complex)
        mps /= np.linalg.norm(mps, axis=0, keepdims=True)
        weights = sp.randn([n]) >= 0

        A = sp.linop.Multiply(shape, weights**0.5) * linop.Sense(mps)
        F = sp.linop.FFT([n])

        p_expected = np.zeros(n, np.complex)
        for i in range(n):
            if weights[i]:
                x = np.zeros(n, np.complex)
                x[i] = 1.0
                p_expected[i] = 1 / F(A.H(A(F.H(x))))[i]

        p = precond.circulant_precond(mps, weights=weights)
        npt.assert_allclose(p[weights == 1], p_expected[weights == 1])

    def test_circulant_precond_noncart(self):
        nc = 4
        n = 10
        shape = [nc, n]
        mps = np.ones(shape, dtype=np.complex)
        mps /= np.linalg.norm(mps, axis=0, keepdims=True)
        coord = sp.randn([n, 1], dtype=np.float)

        A = linop.Sense(mps, coord=coord)
        F = sp.linop.FFT([n])

        p_expected = np.zeros(n, np.complex)
        for i in range(n):
            x = np.zeros(n, np.complex)
            x[i] = 1.0
            p_expected[i] = 1 / F(A.H(A(F.H(x))))[i]

        p = precond.circulant_precond(mps, coord=coord)
        npt.assert_allclose(p, p_expected, atol=1e-1, rtol=1e-1)



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import thresh, config

if config.cupy_enabled:
    import cupy as cp

if __name__ == '__main__':
    unittest.main()


class TestThresh(unittest.TestCase):

    def test_l2_proj(self):
        x = np.ones(5)
        y = np.full(5, 1 / 5**0.5)
        npt.assert_allclose(thresh.l2_proj(1, x), y)

        x = np.ones(5)
        y = np.ones(5)
        npt.assert_allclose(thresh.l2_proj(5**0.5, x), y)

        x = np.ones(5)
        y = np.ones(5)
        npt.assert_allclose(thresh.l2_proj(10, x), y)

    def test_soft_thresh(self):
        x = np.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
        y = np.array([-1, -0.5, 0, 0, 0, 0, 0, 0.5, 1])

        npt.assert_allclose(thresh.soft_thresh(1, x), y)

    def test_hard_thresh(self):
        x = np.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
        y = np.array([-2, -1.5, 0, 0, 0, 0, 0, 1.5, 2])

        npt.assert_allclose(thresh.hard_thresh(1, x), y)

    if config.cupy_enabled:

        def test_soft_thresh_cuda(self):
            x = cp.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
            y = cp.array([-1, -0.5, 0, 0, 0, 0, 0, 0.5, 1])
            lamda = cp.array([1.0])

            cp.testing.assert_allclose(thresh.soft_thresh(lamda, x), y)

        def test_hard_thresh_cuda(self):
            x = cp.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
            y = cp.array([-2, -1.5, 0, 0, 0, 0, 0, 1.5, 2])
            lamda = cp.array([1.0])

            cp.testing.assert_allclose(thresh.hard_thresh(lamda, x), y)



# -*- coding: utf-8 -*-
"""This module contains plotting functions based on matplotlib
for image, line, and scatter plots.

A feature of these plotting functions is that
they can be controlled using only hotkeys
so the user does not need to move away from the keyboard.

Given an array ``x``, an example usage is:

    >>> ImagePlot(x)
    >>> LinePlot(x)
    >>> ScatterPlot(x)

"""
import os
import uuid
import subprocess
import datetime
import numpy as np
import sigpy as sp


__all__ = ['ImagePlot', 'LinePlot', 'ScatterPlot']


class ImagePlot(object):
    """Plot array as image.

    Keyword Args:
        x/y: select current dimension as x and y dimension.
        t: swap between x and y axis.
        z: toggle current dimension as z dimension.
        c: toggle current dimension as color channel.
            Only works if current dimension is of length 3.
        left/right: increment/decrement current dimension
        up/down: flip axis when current dimension is x or y.
            Otherwise increment/decrement slice at current dimension.
        h: toggle hide all labels, titles and axes.
        m: magnitude mode. Renormalizes when pressed each time.
        p: phase mode. Renormalizes when pressed each time.
        r: real mode. Renormalizes when pressed each time.
        i: imaginary mode. Renormalizes when pressed each time.
        l: log mode. Renormalizes when pressed each time.
        [: decreases brightness. Shifts window level up by 10% of window width.
        ]: increases brightness.
            Shifts window level down by 10% of window width.
        {: decreases contrast. Scale window width by 0.9.
        }: increases contrast. Scale window width by 1.1.
        s: save as png.
        g: save as gif by traversing current dimension.
        v: save as mp4 by traversing current dimension.
        q: refresh if file given.
        0-9: enter slice number.
        enter: Set current dimension as slice number.

    """

    def __init__(
            self,
            im,
            x=-1,
            y=-2,
            z=None,
            c=None,
            hide=False,
            mode='m',
            title='',
            interpolation='lanczos',
            save_basename='Figure',
            fps=10,
            filename=None):
        if im.ndim < 2:
            raise TypeError(
                'Image dimension must at least be two, got {im_ndim}'.format(
                    im_ndim=im.ndim))
        import matplotlib.pyplot as plt
        self.axim = None
        self.im = im
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.shape = self.im.shape
        self.ndim = self.im.ndim
        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.x = x % self.ndim
        self.y = y % self.ndim
        self.z = z % self.ndim if z is not None else None
        self.c = c % self.ndim if c is not None else None
        self.d = max(self.ndim - 3, 0)
        self.hide = hide
        self.title = title
        self.interpolation = interpolation
        self.mode = mode
        self.entering_slice = False
        self.vmin = None
        self.vmax = None
        self.save_basename = save_basename
        self.fps = fps
        self.filename = filename

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id)
        self.fig.canvas.mpl_connect('key_press_event', self.key_press)
        self.update_axes()
        self.update_image()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == 'up':
            if self.d not in [self.x, self.y, self.z, self.c]:
                self.slices[self.d] = (
                    self.slices[self.d] + 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 'down':
            if self.d not in [self.x, self.y, self.z, self.c]:
                self.slices[self.d] = (
                    self.slices[self.d] - 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 'left':
            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'right':
            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'x' and self.d not in [self.x, self.z, self.c]:
            if self.d == self.y:
                self.x, self.y = self.y, self.x
            else:
                self.x = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 'y' and self.d not in [self.y, self.z, self.c]:
            if self.d == self.x:
                self.x, self.y = self.y, self.x
            else:
                self.y = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 'z' and self.d not in [self.x, self.y, self.c]:
            if self.d == self.z:
                self.z = None
            else:
                self.z = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif (event.key == 'c' and
              self.d not in [self.x, self.y, self.z] and
              self.shape[self.d] == 3):

            if self.d == self.c:
                self.c = None
            else:
                self.c = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 't':
            self.x, self.y = self.y, self.x

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 'h':
            self.hide = not self.hide

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'f':
            self.fig.canvas.manager.full_screen_toggle()

        elif event.key == 'q':
            if self.filename:
                self.im = np.load(self.filename)
            self.vmin = None
            self.vmax = None
            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == ']':
            width = self.vmax - self.vmin
            self.vmin -= width * 0.1
            self.vmax -= width * 0.1

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == '[':
            width = self.vmax - self.vmin
            self.vmin += width * 0.1
            self.vmax += width * 0.1

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == '}':
            width = self.vmax - self.vmin
            center = (self.vmax + self.vmin) / 2
            self.vmin = center - width * 1.1 / 2
            self.vmax = center + width * 1.1 / 2

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == '{':
            width = self.vmax - self.vmin
            center = (self.vmax + self.vmin) / 2
            self.vmin = center - width * 0.9 / 2
            self.vmax = center + width * 0.9 / 2

            self.update_image()
            self.fig.canvas.draw()

        elif event.key in ['m', 'p', 'r', 'i', 'l']:
            self.vmin = None
            self.vmax = None
            self.mode = event.key

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == 's':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.png')
            self.fig.savefig(filename, transparent=True, format='png',
                             bbox_inches='tight', pad_inches=0)

        elif event.key == 'g':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.gif')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_image()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            'palettegen',
                            '{} palette.png'.format(temp_basename)])
            subprocess.run(['ffmpeg', '-f', 'image2',
                            '-s', '{}x{}'.format(
                                int(bbox.width * self.fig.dpi),
                                int(bbox.height * self.fig.dpi)),
                            '-r', str(self.fps),
                            '-i', '{} %05d.png'.format(temp_basename),
                            '-i', '{} palette.png'.format(temp_basename),
                            '-lavfi', 'paletteuse', filename])

            os.remove('{} palette.png'.format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))

        elif event.key == 'v':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.mp4')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_image()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            "scale=trunc(iw/2)*2:trunc(ih/2)*2",
                            '-vcodec',
                            'libx264',
                            '-pix_fmt',
                            'yuv420p',
                            filename])

            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))

        elif (event.key in ['0', '1', '2', '3', '4',
                            '5', '6', '7', '8', '9', 'backspace'] and
              self.d not in [self.x, self.y, self.z, self.c]):

            if self.entering_slice:
                if event.key == 'backspace':
                    if self.entered_slice < 10:
                        self.entering_slice = False
                    else:
                        self.entered_slice //= 10
                else:
                    self.entered_slice = self.entered_slice * \
                        10 + int(event.key)
            else:
                self.entering_slice = True
                self.entered_slice = int(event.key)

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'enter' and self.entering_slice:
            self.entering_slice = False
            if self.entered_slice < self.shape[self.d]:
                self.slices[self.d] = self.entered_slice

                self.update_image()

            self.update_axes()
            self.fig.canvas.draw()

        else:
            return

    def update_image(self):
        # Extract slice.
        idx = []
        for i in range(self.ndim):
            if i in [self.x, self.y, self.z, self.c]:
                idx.append(slice(None, None, self.flips[i]))
            else:
                idx.append(self.slices[i])

        idx = tuple(idx)
        imv = sp.to_device(self.im[idx])

        # Transpose to have [z, y, x, c].
        imv_dims = [self.y, self.x]
        if self.z is not None:
            imv_dims = [self.z] + imv_dims

        if self.c is not None:
            imv_dims = imv_dims + [self.c]

        imv = np.transpose(imv, np.argsort(np.argsort(imv_dims)))
        imv = array_to_image(imv, color=self.c is not None)

        if self.mode == 'm':
            imv = np.abs(imv)
        elif self.mode == 'p':
            imv = np.angle(imv)
        elif self.mode == 'r':
            imv = np.real(imv)
        elif self.mode == 'i':
            imv = np.imag(imv)
        elif self.mode == 'l':
            imv = np.abs(imv)
            imv = np.log(imv, out=np.ones_like(imv) * -31, where=imv != 0)

        if self.vmin is None:
            self.vmin = imv.min()

        if self.vmax is None:
            self.vmax = imv.max()

        if self.axim is None:
            self.axim = self.ax.imshow(
                imv,
                vmin=self.vmin,
                vmax=self.vmax,
                cmap='gray',
                origin='lower',
                interpolation=self.interpolation,
                aspect=1.0,
                extent=[
                    0,
                    imv.shape[1],
                    0,
                    imv.shape[0]])

        else:
            self.axim.set_data(imv)
            self.axim.set_extent([0, imv.shape[1], 0, imv.shape[0]])
            self.axim.set_clim(self.vmin, self.vmax)

    def update_axes(self):
        if not self.hide:
            caption = '['
            for i in range(self.ndim):

                if i == self.d:
                    caption += '['
                else:
                    caption += ' '

                if (self.flips[i] == -1 and (i == self.x or
                                             i == self.y or
                                             i == self.z)):
                    caption += '-'

                if i == self.x:
                    caption += 'x'
                elif i == self.y:
                    caption += 'y'
                elif i == self.z:
                    caption += 'z'
                elif i == self.c:
                    caption += 'c'
                elif i == self.d and self.entering_slice:
                    caption += str(self.entered_slice) + '_'
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += ']'
                else:
                    caption += ' '
            caption += ']'

            self.ax.set_title(caption)
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title('')
            self.fig.suptitle('')
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)


def mosaic_shape(batch):

    mshape = [int(batch**0.5), batch // int(batch**0.5)]

    while (sp.prod(mshape) < batch):
        mshape[1] += 1

    if (mshape[0] - 1) * (mshape[1] + 1) == batch:
        mshape[0] -= 1
        mshape[1] += 1

    return tuple(mshape)


def array_to_image(arr, color=False):
    """
    Flattens all dimensions except the last two

    """
    if color:
        arr = np.divide(arr, np.abs(arr).max(),
                        out=np.zeros_like(arr), where=arr != 0)

    if arr.ndim == 2:
        return arr
    elif color and arr.ndim == 3:
        return arr

    if color:
        ndim = 3
    else:
        ndim = 2

    arr = sp.resize(arr, arr.shape[:-2] +
                    (arr.shape[-2] + 2, arr.shape[-1] + 2))
    shape = arr.shape
    batch = sp.prod(shape[:-ndim])
    mshape = mosaic_shape(batch)

    if sp.prod(mshape) == batch:
        img = arr.reshape((batch, ) + shape[-ndim:])
    else:
        img = np.zeros((sp.prod(mshape), ) + shape[-ndim:], dtype=arr.dtype)
        img[:batch, ...] = arr.reshape((batch, ) + shape[-ndim:])

    img = img.reshape(mshape + shape[-ndim:])
    if color:
        img = np.transpose(img, (0, 2, 1, 3, 4))
        img = img.reshape(
            (shape[-3] * mshape[-2], shape[-2] * mshape[-1], shape[-1]))
    else:
        img = np.transpose(img, (0, 2, 1, 3))
        img = img.reshape((shape[-2] * mshape[-2], shape[-1] * mshape[-1]))

    return img


class LinePlot(object):
    """Plot array as lines.

    Keyword Args:
        x: select current dimension as x
        left/right: increment/decrement current dimension
        up/down: flip axis when current dimension is x or y
            otherwise increment/decrement slice at current dimension
        h: toggle hide all labels, titles and axes
        m: magnitude mode
        p: phase mode
        r: real mode
        i: imaginary mode
        l: log mode
        s: save as png.
        g: save as gif by traversing current dimension.
        v: save as mp4 by traversing current dimension.
    """

    def __init__(self, arr, x=-1, hide=False, mode='m', title='',
                 save_basename='Figure', fps=10):
        import matplotlib.pyplot as plt

        self.arr = arr
        self.axarr = None

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.shape = self.arr.shape
        self.ndim = self.arr.ndim
        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.x = x % self.ndim
        self.d = max(self.ndim - 3, 0)
        self.hide = hide
        self.title = title
        self.mode = mode
        self.save_basename = save_basename
        self.fps = fps

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id)
        self.fig.canvas.mpl_connect('key_press_event', self.key_press)
        self.update_axes()
        self.update_line()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == 'up':
            if self.d != self.x:
                self.slices[self.d] = (
                    self.slices[self.d] + 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == 'down':

            if self.d != self.x:
                self.slices[self.d] = (
                    self.slices[self.d] - 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == 'left':

            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'right':

            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'x' and self.d != self.x:

            self.x = self.d

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == 'h':
            self.hide = not self.hide

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'f':
            self.fig.canvas.manager.full_screen_toggle()

        elif (event.key == 'm' or event.key == 'p' or
              event.key == 'r' or event.key == 'i' or event.key == 'l'):

            self.mode = event.key

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == 's':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.png')
            self.fig.savefig(filename, transparent=True, format='png',
                             bbox_inches='tight', pad_inches=0)

        elif event.key == 'g':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.gif')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_line()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            'palettegen',
                            '{} palette.png'.format(temp_basename)])
            subprocess.run(['ffmpeg', '-f', 'image2',
                            '-s', '{}x{}'.format(
                                int(bbox.width * self.fig.dpi),
                                int(bbox.height * self.fig.dpi)),
                            '-r', str(self.fps),
                            '-i', '{} %05d.png'.format(temp_basename),
                            '-i', '{} palette.png'.format(temp_basename),
                            '-lavfi', 'paletteuse', filename])

            os.remove('{} palette.png'.format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))

        elif event.key == 'v':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %h.%M.%S %p.mp4')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_line()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            "scale=trunc(iw/2)*2:trunc(ih/2)*2",
                            '-vcodec',
                            'libx264',
                            '-pix_fmt',
                            'yuv420p',
                            filename])

            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))
        else:
            return

        return

    def update_line(self):

        order = ([i for i in range(self.ndim)
                  if i != self.x] + [self.x])
        idx = tuple([self.slices[i] for i in order[:-1]] +
                    [slice(None, None, self.flips[self.x])])

        arrv = self.arr.transpose(order)[idx]

        if self.mode == 'm':
            arrv = np.abs(arrv)
        elif self.mode == 'a':
            arrv = np.angle(arrv)
        elif self.mode == 'r':
            arrv = np.real(arrv)
        elif self.mode == 'i':
            arrv = np.imag(arrv)
        elif self.mode == 'l':
            eps = 1e-31
            arrv = np.log(np.abs(arrv) + eps)

        if self.axarr is None:
            self.axarr = self.ax.plot(arrv)[0]

        else:
            self.axarr.set_xdata(np.arange(len(arrv)))
            self.axarr.set_ydata(arrv)
            self.ax.relim()
            self.ax.autoscale_view()

    def update_axes(self):

        if not self.hide:
            caption = 'Slice: ['
            for i in range(self.ndim):

                if i == self.d:
                    caption += '['
                else:
                    caption += ' '

                if self.flips[i] == -1 and i == self.x:
                    caption += '-'

                if i == self.x:
                    caption += 'x'
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += ']'
                else:
                    caption += ' '
            caption += ']'

            self.ax.set_title(caption)
            self.ax.axis('on')
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title('')
            self.fig.suptitle('')
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)


class ScatterPlot(object):
    """Plot array as scatter.

    Keyword Args:
        z: toggle current dimension as z dimension
        left/right: increment/decrement current dimension
        up/down: flip axis when current dimension is x or y
            otherwise increment/decrement slice at current dimension
        h: toggle hide all labels, titles and axes
        m: magnitude mode
        p: phase mode
        r: real mode
        i: imaginary mode
        l: log mode
    """

    def __init__(
            self,
            coord,
            data=None,
            z=None,
            hide=False,
            mode='m',
            title='',
            save_basename='Figure',
            fps=10):
        import matplotlib.pyplot as plt

        self.coord = coord
        assert coord.shape[-1] == 2
        if data is None:
            self.data = np.ones(coord.shape[:-1])
        else:
            self.data = data

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_facecolor('k')

        for c, d in zip(coord.shape[:-1], self.data.shape[-coord.ndim + 1:]):
            assert c == d

        self.ndim = self.data.ndim - self.coord.ndim + 1
        self.shape = self.data.shape[:self.ndim]

        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.z = z % self.ndim if z is not None else None
        self.d = 0
        self.hide = hide
        self.title = title
        self.mode = mode
        self.axsc = None
        self.entering_slice = False
        self.save_basename = save_basename
        self.fps = fps
        self.vmin = None
        self.vmax = None

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id)
        self.fig.canvas.mpl_connect('key_press_event', self.key_press)
        self.update_axes()
        self.update_data()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == 'up':
            if self.d != self.z:
                self.slices[self.d] = (
                    self.slices[self.d] + 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == 'down':
            if self.d != self.z:
                self.slices[self.d] = (
                    self.slices[self.d] - 1) % self.shape[self.d]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == 'left':
            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'right':
            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        # elif event.key == 'z':
        #     if self.d == self.z:
        #         self.z = None
        #     else:
        #         self.z = self.d

        #     self.update_axes()
        #     self.update_data()
        #     self.fig.canvas.draw()

        elif event.key == 'h':
            self.hide = not self.hide

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'f':
            self.fig.canvas.manager.full_screen_toggle()

        elif (event.key == 'm' or event.key == 'p' or
              event.key == 'r' or event.key == 'i' or event.key == 'l'):
            self.mode = event.key
            self.vmin = None
            self.vmax = None

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == 's':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.png')
            self.fig.savefig(filename, transparent=True, format='png',
                             bbox_inches='tight', pad_inches=0)

        elif event.key == 'g':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %I.%M.%S %p.gif')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_data()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            'palettegen',
                            '{} palette.png'.format(temp_basename)])
            subprocess.run(['ffmpeg', '-f', 'image2',
                            '-s', '{}x{}'.format(
                                int(bbox.width * self.fig.dpi),
                                int(bbox.height * self.fig.dpi)),
                            '-r', str(self.fps),
                            '-i', '{} %05d.png'.format(temp_basename),
                            '-i', '{} palette.png'.format(temp_basename),
                            '-lavfi', 'paletteuse', filename])

            os.remove('{} palette.png'.format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))

        elif event.key == 'v':
            filename = self.save_basename + \
                datetime.datetime.now().strftime(
                    ' %Y-%m-%d at %h.%M.%S %p.mp4')
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_data()
                self.fig.canvas.draw()
                self.fig.savefig('{} {:05d}.png'.format(temp_basename, i),
                                 format='png', bbox_inches=bbox, pad_inches=0)

            subprocess.run(['ffmpeg',
                            '-f',
                            'image2',
                            '-s',
                            '{}x{}'.format(int(bbox.width * self.fig.dpi),
                                           int(bbox.height * self.fig.dpi)),
                            '-r',
                            str(self.fps),
                            '-i',
                            '{} %05d.png'.format(temp_basename),
                            '-vf',
                            "scale=trunc(iw/2)*2:trunc(ih/2)*2",
                            '-vcodec',
                            'libx264',
                            '-pix_fmt',
                            'yuv420p',
                            filename])

            for i in range(self.shape[self.d]):
                os.remove('{} {:05d}.png'.format(temp_basename, i))

        elif (event.key in ['0', '1', '2', '3', '4', '5',
                            '6', '7', '8', '9', 'backspace'] and
              self.d != self.z):

            if self.entering_slice:
                if event.key == 'backspace':
                    if self.entered_slice < 10:
                        self.entering_slice = False
                    else:
                        self.entered_slice //= 10
                else:
                    self.entered_slice = self.entered_slice * \
                        10 + int(event.key)
            else:
                self.entering_slice = True
                self.entered_slice = int(event.key)

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == 'enter' and self.entering_slice:
            self.entering_slice = False
            if self.entered_slice < self.shape[self.d]:
                self.slices[self.d] = self.entered_slice

                self.update_data()

            self.update_axes()
            self.fig.canvas.draw()

        else:
            return

    def update_data(self):

        idx = []
        for i in range(self.ndim):
            if i == self.z:
                idx.append(slice(None, None, self.flips[i]))
            else:
                idx.append(self.slices[i])

        idx = tuple(idx)
        if idx:
            datav = sp.to_device(self.data[idx])
        else:
            datav = sp.to_device(self.data)

        # if self.z is not None:
        #     datav_dims = [self.z] + datav_dims
        coordv = sp.to_device(self.coord)

        if self.mode == 'm':
            datav = np.abs(datav)
        elif self.mode == 'p':
            datav = np.angle(datav)
        elif self.mode == 'r':
            datav = np.real(datav)
        elif self.mode == 'i':
            datav = np.imag(datav)
        elif self.mode == 'l':
            eps = 1e-31
            datav = np.log(np.abs(datav) + eps)

        datav = datav.ravel()
        if self.vmin is None:
            if datav.min() == datav.max():
                self.vmin = 0
            else:
                self.vmin = datav.min()

        if self.vmax is None:
            self.vmax = datav.max()

        if self.axsc is None:
            self.axsc = self.ax.scatter(
                coordv[..., 0].ravel(), coordv[..., 1].ravel(), c=datav,
                s=1, linewidths=0, cmap='gray',
                vmin=self.vmin, vmax=self.vmax,
            )

        else:
            self.axsc.set_offsets(coordv.T.reshape([-1, 2]))
            self.axsc.set_color(datav)

    def update_axes(self):

        if not self.hide:
            caption = '['
            for i in range(self.ndim):

                if i == self.d:
                    caption += '['
                else:
                    caption += ' '

                if (self.flips[i] == -1 and i == self.z):
                    caption += '-'

                if i == self.z:
                    caption += 'z'
                elif i == self.d and self.entering_slice:
                    caption += str(self.entered_slice) + '_'
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += ']'
                else:
                    caption += ' '
            caption += ']'

            self.ax.set_title(caption)
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title('')
            self.fig.suptitle('')
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)



import numpy as np

__all__ = ['shepp_logan']


def shepp_logan(shape, dtype=np.complex):
    """Generates a Shepp Logan phantom with a given shape and dtype.

    Args:
        shape (tuple of ints): shape, can be of length 2 or 3.
        dtype (Dtype): data type.

    Returns:
        array.

    """
    return phantom(shape, sl_amps, sl_scales, sl_offsets, sl_angles, dtype)


sl_amps = [1, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

sl_scales = [[.6900, .920, .810],  # white big
             [.6624, .874, .780],  # gray big
             [.1100, .310, .220],  # right black
             [.1600, .410, .280],  # left black
             [.2100, .250, .410],  # gray center blob
             [.0460, .046, .050],
             [.0460, .046, .050],
             [.0460, .046, .050],  # left small dot
             [.0230, .023, .020],  # mid small dot
             [.0230, .023, .020]]

sl_offsets = [[0., 0., 0],
              [0., -.0184, 0],
              [.22, 0., 0],
              [-.22, 0., 0],
              [0., .35, -.15],
              [0., .1, .25],
              [0., -.1, .25],
              [-.08, -.605, 0],
              [0., -.606, 0],
              [.06, -.605, 0]]

sl_angles = [[0, 0, 0],
             [0, 0, 0],
             [-18, 0, 10],
             [18, 0, 10],
             [0, 0, 0],
             [0, 0, 0],
             [0, 0, 0],
             [0, 0, 0],
             [0, 0, 0],
             [0, 0, 0]]


def phantom(shape, amps, scales, offsets, angles, dtype):
    """
    Generate a cube of given shape using a list of ellipsoid
    parameters.
    """

    if len(shape) == 2:

        ndim = 2
        shape = (1, shape[-2], shape[-1])

    elif len(shape) == 3:

        ndim = 3

    else:

        raise ValueError('Incorrect dimension')

    out = np.zeros(shape, dtype=dtype)

    z, y, x = np.mgrid[-(shape[-3] // 2):((shape[-3] + 1) // 2),
                       -(shape[-2] // 2):((shape[-2] + 1) // 2),
                       -(shape[-1] // 2):((shape[-1] + 1) // 2)]

    coords = np.stack((x.ravel() / shape[-1] * 2,
                       y.ravel() / shape[-2] * 2,
                       z.ravel() / shape[-3] * 2))

    for amp, scale, offset, angle in zip(amps, scales, offsets, angles):

        ellipsoid(amp, scale, offset, angle, coords, out)

    if ndim == 2:

        return out[0, :, :]

    else:

        return out


def ellipsoid(amp, scale, offset, angle, coords, out):
    """
    Generate a cube containing an ellipsoid defined by its parameters.
    If out is given, fills the given cube instead of creating a new
    one.
    """
    R = rotation_matrix(angle)
    coords = (np.matmul(R, coords) - np.reshape(offset, (3, 1))) / \
        np.reshape(scale, (3, 1))

    r2 = np.sum(coords ** 2, axis=0).reshape(out.shape)

    out[r2 <= 1] += amp


def rotation_matrix(angle):
    cphi = np.cos(np.radians(angle[0]))
    sphi = np.sin(np.radians(angle[0]))
    ctheta = np.cos(np.radians(angle[1]))
    stheta = np.sin(np.radians(angle[1]))
    cpsi = np.cos(np.radians(angle[2]))
    spsi = np.sin(np.radians(angle[2]))
    alpha = [[cpsi * cphi - ctheta * sphi * spsi,
              cpsi * sphi + ctheta * cphi * spsi,
              spsi * stheta],
             [-spsi * cphi - ctheta * sphi * cpsi,
              -spsi * sphi + ctheta * cphi * cpsi,
              cpsi * stheta],
             [stheta * sphi,
              -stheta * cphi,
              ctheta]]
    return np.array(alpha)



# -*- coding: utf-8 -*-
"""This module provides functions for interoperability
between sigpy and pytorch.
"""
import numpy as np
from sigpy import backend, config

__all__ = []
if config.pytorch_enabled:  # pragma: no cover
    __all__ += ['to_pytorch', 'from_pytorch', 'to_pytorch_function']

    def to_pytorch(array, requires_grad=True):
        """Zero-copy conversion from numpy/cupy array to pytorch tensor.

        For complex array input, returns a tensor with shape + [2],
        where tensor[..., 0] and tensor[..., 1] represent the real
        and imaginary.

        Args:
            array (numpy/cupy array): input.

        Returns:
            PyTorch tensor.

        """
        import torch
        from torch.utils.dlpack import from_dlpack

        device = backend.get_device(array)
        if not np.issubdtype(array.dtype, np.floating):
            with device:
                shape = array.shape
                array = array.view(dtype=array.real.dtype)
                array = array.reshape(shape + (2, ))

        if device == backend.cpu_device:
            tensor = torch.from_numpy(array)
        else:
            tensor = from_dlpack(array.toDlpack())

        tensor.requires_grad = requires_grad
        return tensor

    def from_pytorch(tensor, iscomplex=False):
        """Zero-copy conversion from pytorch tensor to numpy/cupy array.

        If iscomplex, then tensor must have the last dimension as 2,
        and the output will be viewed as a complex valued array.

        Args:
            tensor (PyTorch tensor): input.
            iscomplex (bool): whether input represents complex valued tensor.

        Returns:
            Numpy/cupy array.

        """
        from torch.utils.dlpack import to_dlpack

        device = tensor.device
        if device.type == 'cpu':
            output = tensor.detach().numpy()
        else:
            if config.cupy_enabled:
                import cupy as cp
                output = cp.fromDlpack(to_dlpack(tensor))
            else:
                raise TypeError('CuPy not installed, '
                                'but trying to convert GPU PyTorch Tensor.')

        if iscomplex:
            if output.shape[-1] != 2:
                raise ValueError('shape[-1] must be 2 when iscomplex is '
                                 'specified, but got {}'.format(output.shape))

            with backend.get_device(output):
                if output.dtype == np.float32:
                    output = output.view(np.complex64)
                elif output.dtype == np.float64:
                    output = output.view(np.complex128)

                output = output.reshape(output.shape[:-1])

        return output

    def to_pytorch_function(linop,
                            input_iscomplex=False,
                            output_iscomplex=False):
        """Convert SigPy Linop to PyTorch Function.

        The returned function can be treated as a native
        pytorch function performing the linop operator.
        The function can be backpropagated, applied on GPU arrays,
        and has minimal overhead as the underlying arrays
        are shared without copying.
        For complex valued input/output, the appropriate options
        should be set when calling the function.

        Args:
            linop (Linop): linear operator to be converted.
            input_iscomplex (bool): whether the PyTorch input
                represents complex tensor.
            output_iscomplex (bool): whether the PyTorch output
                represents complex tensor.

        Returns:
            torch.autograd.Function: equivalent PyTorch Function.

        """
        import torch

        class LinopFunction(torch.autograd.Function):
            @staticmethod
            def forward(ctx, input):
                return to_pytorch(linop(from_pytorch(
                    input, iscomplex=input_iscomplex)))

            @staticmethod
            def backward(ctx, grad_output):
                return to_pytorch(linop.H(from_pytorch(
                    grad_output, iscomplex=output_iscomplex)))

        return LinopFunction



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import fourier, util

if __name__ == '__main__':
    unittest.main()


class TestFourier(unittest.TestCase):

    def test_fft(self):
        input = np.array([0, 1, 0], dtype=np.complex)
        npt.assert_allclose(fourier.fft(input),
                            np.ones(3) / 3**0.5, atol=1e-5)

        input = np.array([1, 1, 1], dtype=np.complex)
        npt.assert_allclose(fourier.fft(input),
                            [0, 3**0.5, 0], atol=1e-5)

        input = util.randn([4, 5, 6])
        npt.assert_allclose(fourier.fft(input),
                            np.fft.fftshift(np.fft.fftn(
                                np.fft.ifftshift(input), norm='ortho')),
                            atol=1e-5)

        input = np.array([0, 1, 0], dtype=np.complex)
        npt.assert_allclose(fourier.fft(input, oshape=[5]),
                            np.ones(5) / 5**0.5, atol=1e-5)

    def test_fft_dtype(self):

        for dtype in [np.complex64, np.complex128]:
            input = np.array([0, 1, 0], dtype=dtype)
            output = fourier.fft(input)

            assert output.dtype == dtype

    def test_ifft(self):
        input = np.array([0, 1, 0], dtype=np.complex)
        npt.assert_allclose(fourier.ifft(input),
                            np.ones(3) / 3 ** 0.5, atol=1e-5)

        input = np.array([1, 1, 1], dtype=np.complex)
        npt.assert_allclose(fourier.ifft(input),
                            [0, 3**0.5, 0], atol=1e-5)

        input = util.randn([4, 5, 6])
        npt.assert_allclose(fourier.ifft(input),
                            np.fft.fftshift(np.fft.ifftn(
                                np.fft.ifftshift(input), norm='ortho')),
                            atol=1e-5)

        input = np.array([0, 1, 0], dtype=np.complex)
        npt.assert_allclose(fourier.ifft(input, oshape=[5]),
                            np.ones(5) / 5**0.5, atol=1e-5)

    def test_nufft(self):

        # Check deltas
        input = np.array([0, 1, 0], np.complex)  # delta
        coord = np.array([[-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft(input, coord),
                            np.array([1.0, 1.0, 1.0]) / (3**0.5),
                            atol=0.01, rtol=0.01)

        input = np.array([0, 0, 1, 0], np.complex)  # delta
        coord = np.array([[-2], [-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft(input, coord),
                            np.array([1.0, 1.0, 1.0, 1.0]) / (4**0.5),
                            atol=0.01, rtol=0.01)

        input = np.array([0, 0, 1, 0, 0], np.complex)  # delta
        coord = np.array([[-2], [-1], [0], [1], [2]], np.float)

        npt.assert_allclose(fourier.nufft(input, coord),
                            np.ones(5) / (5**0.5),
                            atol=0.01, rtol=0.01)

        # Check shifted delta
        input = np.array([0, 0, 1], np.complex)  # shifted delta
        coord = np.array([[-1], [0], [1]], np.float)

        w = np.exp(-1j * 2.0 * np.pi / 3.0)
        npt.assert_allclose(fourier.nufft(input, coord),
                            np.array([w.conjugate(), 1.0, w]) / (3**0.5),
                            atol=0.01, rtol=0.01)

        input = np.array([0, 0, 0, 1], np.complex)  # delta
        coord = np.array([[-2], [-1], [0], [1]], np.float)

        w = np.exp(-1j * 2.0 * np.pi / 4.0)
        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.array(
                [w.conjugate()**2, w.conjugate(), 1.0, w]) / (4**0.5),
            atol=0.01, rtol=0.01)

    def test_nufft_nd(self):

        for ndim in range(3):
            input = np.array([[0], [1], [0]], np.complex)
            coord = np.array([[-1, 0],
                              [0, 0],
                              [1, 0]], np.float)

            npt.assert_allclose(fourier.nufft(input, coord),
                                np.array([1.0, 1.0, 1.0]) / 3**0.5,
                                atol=0.01, rtol=0.01)

    def test_nufft_adjoint(self):

        # Check deltas
        oshape = [3]
        input = np.array([1.0, 1.0, 1.0], dtype=np.complex) / 3**0.5
        coord = np.array([[-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([0, 1, 0]),
                            atol=0.01, rtol=0.01)

        oshape = [4]
        input = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.complex) / 4**0.5
        coord = np.array([[-2], [-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([0, 0, 1, 0], np.complex),
                            atol=0.01, rtol=0.01)

        oshape = [5]
        input = np.ones(5, dtype=np.complex) / 5**0.5
        coord = np.array([[-2], [-1], [0], [1], [2]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([0, 0, 1, 0, 0], np.complex),
                            atol=0.01, rtol=0.01)

        # Check shifted delta
        oshape = [3]
        w = np.exp(-1j * 2.0 * np.pi / 3.0)
        input = np.array([w.conjugate(), 1.0, w]) / 3**0.5
        coord = np.array([[-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([0, 0, 1], np.complex),
                            atol=0.01, rtol=0.01)

        oshape = [4]
        w = np.exp(-1j * 2.0 * np.pi / 4.0)
        input = np.array([w.conjugate()**2, w.conjugate(), 1.0, w]) / 4**0.5
        coord = np.array([[-2], [-1], [0], [1]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([0, 0, 0, 1], np.complex),
                            atol=0.01, rtol=0.01)

    def test_nufft_adjoint_nd(self):

        oshape = [3, 1]

        input = np.array([1.0, 1.0, 1.0], dtype=np.complex) / 3**0.5
        coord = np.array([[-1, 0],
                          [0, 0],
                          [1, 0]], np.float)

        npt.assert_allclose(fourier.nufft_adjoint(input, coord, oshape),
                            np.array([[0], [1], [0]], np.complex),
                            atol=0.01, rtol=0.01)

    def test_nufft_normal(self):

        # Check delta
        oshape = [3]
        input = np.array([0.0, 1.0, 0.0], dtype=np.complex)
        coord = np.array([[-1], [0], [1]], np.float)

        npt.assert_allclose(
            fourier.nufft_adjoint(fourier.nufft(
                input, coord), coord, oshape), np.array([0, 1, 0]),
            atol=0.01, rtol=0.01)

        # Check delta scale
        oshape = [3]
        input = np.array([0.0, 1.0, 0.0], dtype=np.complex)
        coord = np.array([[-1], [-0.5], [0], [0.5], [1]], np.float)

        npt.assert_allclose(
            fourier.nufft_adjoint(
                fourier.nufft(
                    input,
                    coord),
                coord,
                oshape)[
                len(input) // 2],
            5 / 3,
            atol=0.01,
            rtol=0.01)

    def test_nufft_ndft(self):

        n = 5
        w = np.exp(-1j * 2 * np.pi / n)
        coord = np.array([[-2], [0], [0.1]])
        w2 = w**-2
        w1 = w**0.1
        A = np.array([[w2**-2, w2**-1, 1, w2, w2**2],
                      [1, 1, 1, 1, 1],
                      [w1**-2, w1**-1, 1, w1, w1**2]]) / n**0.5

        for i in range(n):
            input = np.zeros(n, np.complex)
            input[i] = 1
            npt.assert_allclose(A[:, i], fourier.nufft(
                input, coord), atol=0.01, rtol=0.01)

        n = 6
        w = np.exp(-1j * 2 * np.pi / n)
        coord = np.array([[-2], [0], [0.1]])
        w2 = w**-2
        w1 = w**0.1
        A = np.array([[w2**-3, w2**-2, w2**-1, 1, w2, w2**2],
                      [1, 1, 1, 1, 1, 1],
                      [w1**-3, w1**-2, w1**-1, 1, w1, w1**2]]) / n**0.5

        for i in range(n):
            input = np.zeros(n, np.complex)
            input[i] = 1
            npt.assert_allclose(A[:, i], fourier.nufft(
                input, coord), atol=0.01, rtol=0.01)



# -*- coding: utf-8 -*-
"""Interpolation functions.
"""
import numpy as np
import numba as nb

from sigpy import backend, config, util


__all__ = ['interpolate', 'gridding']


def interpolate(input, width, kernel, coord):
    """Interpolation from array to points specified by coordinates.

    Args:
        input (array): Input array of shape [..., ny, nx]
        width (float): Interpolation kernel width.
        kernel (array): Interpolation kernel.
        coord (array): Coordinate array of shape [..., ndim]

    Returns:
        output (array): Output array of coord.shape[:-1]

    """
    ndim = coord.shape[-1]

    batch_shape = input.shape[:-ndim]
    batch_size = util.prod(batch_shape)

    pts_shape = coord.shape[:-1]
    npts = util.prod(pts_shape)

    device = backend.get_device(input)
    xp = device.xp
    isreal = np.issubdtype(input.dtype, np.floating)
    coord = backend.to_device(coord, device)
    kernel = backend.to_device(kernel, device)

    with device:
        input = input.reshape([batch_size] + list(input.shape[-ndim:]))
        coord = coord.reshape([npts, ndim])
        output = xp.zeros([batch_size, npts], dtype=input.dtype)

        _interpolate = _select_interpolate(ndim, npts, device, isreal)
        if device == backend.cpu_device:
            _interpolate(output, input, width, kernel, coord)
        else:  # pragma: no cover
            _interpolate(input, width, kernel, coord, output, size=npts)

        return output.reshape(batch_shape + pts_shape)


def gridding(input, shape, width, kernel, coord):
    """Gridding of points specified by coordinates to array.

    Args:
        input (array): Input array.
        shape (array of ints): Output shape.
        width (float): Interpolation kernel width.
        kernel (array): Interpolation kernel.
        coord (array): Coordinate array of shape [..., ndim]

    Returns:
        output (array): Output array.

    """
    ndim = coord.shape[-1]

    batch_shape = shape[:-ndim]
    batch_size = util.prod(batch_shape)

    pts_shape = coord.shape[:-1]
    npts = util.prod(pts_shape)

    device = backend.get_device(input)
    xp = device.xp
    isreal = np.issubdtype(input.dtype, np.floating)

    with device:
        input = input.reshape([batch_size, npts])
        coord = coord.reshape([npts, ndim])
        output = xp.zeros(
            [batch_size] + list(shape[-ndim:]), dtype=input.dtype)

        _gridding = _select_gridding(ndim, npts, device, isreal)
        if device == backend.cpu_device:
            _gridding(output, input, width, kernel, coord)
        else:  # pragma: no cover
            _gridding(input, width, kernel, coord, output, size=npts)

        return output.reshape(shape)


def _select_interpolate(ndim, npts, device, isreal):
    if ndim == 1:
        if device == backend.cpu_device:
            _interpolate = _interpolate1
        else:  # pragma: no cover
            _interpolate = _interpolate1_cuda
    elif ndim == 2:
        if device == backend.cpu_device:
            _interpolate = _interpolate2
        else:  # pragma: no cover
            _interpolate = _interpolate2_cuda
    elif ndim == 3:
        if device == backend.cpu_device:
            _interpolate = _interpolate3
        else:  # pragma: no cover
            _interpolate = _interpolate3_cuda
    else:
        raise ValueError(
            'Number of dimensions can only be 1, 2 or 3, got {}'.format(ndim))

    return _interpolate


def _select_gridding(ndim, npts, device, isreal):
    if ndim == 1:
        if device == backend.cpu_device:
            _gridding = _gridding1
        else:  # pragma: no cover
            if isreal:
                _gridding = _gridding1_cuda
            else:
                _gridding = _gridding1_cuda_complex
    elif ndim == 2:
        if device == backend.cpu_device:
            _gridding = _gridding2
        else:  # pragma: no cover
            if isreal:
                _gridding = _gridding2_cuda
            else:
                _gridding = _gridding2_cuda_complex
    elif ndim == 3:
        if device == backend.cpu_device:
            _gridding = _gridding3
        else:  # pragma: no cover
            if isreal:
                _gridding = _gridding3_cuda
            else:
                _gridding = _gridding3_cuda_complex
    else:
        raise ValueError(
            'Number of dimensions can only be 1, 2 or 3, got {}'.format(ndim))

    return _gridding


@nb.jit(nopython=True)  # pragma: no cover
def lin_interpolate(kernel, x):
    if x >= 1:
        return 0.0
    n = len(kernel)
    idx = int(x * n)
    frac = x * n - idx

    left = kernel[idx]
    if idx == n - 1:
        right = 0.0
    else:
        right = kernel[idx + 1]
    return (1.0 - frac) * left + frac * right


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _interpolate1(output, input, width, kernel, coord):
    batch_size, nx = input.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx = coord[i, -1]

        x0 = np.ceil(kx - width / 2)
        x1 = np.floor(kx + width / 2)

        for x in range(x0, x1 + 1):

            w = lin_interpolate(kernel, abs(x - kx) / (width / 2))

            for b in range(batch_size):
                output[b, i] += w * input[b, x % nx]

    return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _gridding1(output, input, width, kernel, coord):
    batch_size, nx = output.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx = coord[i, -1]

        x0 = np.ceil(kx - width / 2)
        x1 = np.floor(kx + width / 2)

        for x in range(x0, x1 + 1):

            w = lin_interpolate(kernel, abs(x - kx) / (width / 2))

            for b in range(batch_size):
                output[b, x % nx] += w * input[b, i]

    return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _interpolate2(output, input, width, kernel, coord):

    batch_size, ny, nx = input.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx, ky = coord[i, -1], coord[i, -2]

        x0, y0 = (np.ceil(kx - width / 2),
                  np.ceil(ky - width / 2))

        x1, y1 = (np.floor(kx + width / 2),
                  np.floor(ky + width / 2))

        for y in range(y0, y1 + 1):
            wy = lin_interpolate(kernel, abs(y - ky) / (width / 2))

            for x in range(x0, x1 + 1):
                w = wy * lin_interpolate(kernel, abs(x - kx) / (width / 2))

                for b in range(batch_size):
                    output[b, i] += w * input[b, y % ny, x % nx]

    return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _gridding2(output, input, width, kernel, coord):
    batch_size, ny, nx = output.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx, ky = coord[i, -1], coord[i, -2]

        x0, y0 = (np.ceil(kx - width / 2),
                  np.ceil(ky - width / 2))

        x1, y1 = (np.floor(kx + width / 2),
                  np.floor(ky + width / 2))

        for y in range(y0, y1 + 1):
            wy = lin_interpolate(kernel, abs(y - ky) / (width / 2))

            for x in range(x0, x1 + 1):
                w = wy * lin_interpolate(kernel, abs(x - kx) / (width / 2))

                for b in range(batch_size):
                    output[b, y % ny, x % nx] += w * input[b, i]

    return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _interpolate3(output, input, width, kernel, coord):
    batch_size, nz, ny, nx = input.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx, ky, kz = coord[i, -1], coord[i, -2], coord[i, -3]

        x0, y0, z0 = (np.ceil(kx - width / 2),
                      np.ceil(ky - width / 2),
                      np.ceil(kz - width / 2))

        x1, y1, z1 = (np.floor(kx + width / 2),
                      np.floor(ky + width / 2),
                      np.floor(kz + width / 2))

        for z in range(z0, z1 + 1):
            wz = lin_interpolate(kernel, abs(z - kz) / (width / 2))

            for y in range(y0, y1 + 1):
                wy = wz * lin_interpolate(kernel, abs(y - ky) / (width / 2))

                for x in range(x0, x1 + 1):
                    w = wy * lin_interpolate(kernel, abs(x - kx) / (width / 2))

                    for b in range(batch_size):
                        output[b, i] += w * input[b, z % nz, y % ny, x % nx]

    return output


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _gridding3(output, input, width, kernel, coord):
    batch_size, nz, ny, nx = output.shape
    npts = coord.shape[0]

    for i in range(npts):

        kx, ky, kz = coord[i, -1], coord[i, -2], coord[i, -3]

        x0, y0, z0 = (np.ceil(kx - width / 2),
                      np.ceil(ky - width / 2),
                      np.ceil(kz - width / 2))

        x1, y1, z1 = (np.floor(kx + width / 2),
                      np.floor(ky + width / 2),
                      np.floor(kz + width / 2))

        for z in range(z0, z1 + 1):
            wz = lin_interpolate(kernel, abs(z - kz) / (width / 2))

            for y in range(y0, y1 + 1):
                wy = wz * lin_interpolate(kernel, abs(y - ky) / (width / 2))

                for x in range(x0, x1 + 1):
                    w = wy * lin_interpolate(kernel, abs(x - kx) / (width / 2))

                    for b in range(batch_size):
                        output[b, z % nz, y % ny, x % nx] += w * input[b, i]

    return output


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    lin_interpolate_cuda = """
    __device__ inline S lin_interpolate(const S* kernel, int n, S x) {
        if (x >= 1)
           return 0;
        const int idx = x * n;
        const S frac = x * n - idx;

        const S left = kernel[idx];
        S right = 0;
        if (idx != n - 1)
            right = kernel[idx + 1];

        return (1 - frac) * left + frac * right;
    }
    """
    mod_cuda = """
    __device__ inline int mod(int x, int n) {
        return (x % n + n) % n;
    }
    """

    _interpolate1_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = input.shape()[0];
        const int nx = input.shape()[1];

        const int coord_idx[] = {i, 0};
        const S kx = coord[coord_idx];
        const int x0 = ceil(kx - width / 2.0);
        const int x1 = floor(kx + width / 2.0);

        for (int x = x0; x < x1 + 1; x++) {
            const S w = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) x - kx) / (width / 2.0));
            for (int b = 0; b < batch_size; b++) {
                const int input_idx[] = {b, mod(x, nx)};
                const T v = (T) w * input[input_idx];
                const int output_idx[] = {b, i};
                output[output_idx] += v;
            }
        }
        """,
        name='interpolate1',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding1_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = output.shape()[0];
        const int nx = output.shape()[1];

        const int coord_idx[] = {i, 0};
        const S kx = coord[coord_idx];
        const int x0 = ceil(kx - width / 2.0);
        const int x1 = floor(kx + width / 2.0);

        for (int x = x0; x < x1 + 1; x++) {
            const S w = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) x - kx) / (width / 2.0));
            for (int b = 0; b < batch_size; b++) {
                const int input_idx[] = {b, i};
                const T v = (T) w * input[input_idx];
                const int output_idx[] = {b, mod(x, nx)};
                atomicAdd(&output[output_idx], v);
            }
        }
        """,
        name='gridding1',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding1_cuda_complex = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = output.shape()[0];
        const int nx = output.shape()[1];

        const int coord_idx[] = {i, 0};
        const S kx = coord[coord_idx];
        const int x0 = ceil(kx - width / 2.0);
        const int x1 = floor(kx + width / 2.0);

        for (int x = x0; x < x1 + 1; x++) {
            const S w = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) x - kx) / (width / 2.0));
            for (int b = 0; b < batch_size; b++) {
                const int input_idx[] = {b, i};
                const T v = (T) w * input[input_idx];
                const int output_idx[] = {b, mod(x, nx)};
                atomicAdd(
                    reinterpret_cast<T::value_type*>(
                        &(output[output_idx])), v.real());
                atomicAdd(
                    reinterpret_cast<T::value_type*>(
                        &(output[output_idx])) + 1, v.imag());
            }
        }
        """,
        name='gridding1_complex',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _interpolate2_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = input.shape()[0];
        const int ny = input.shape()[1];
        const int nx = input.shape()[2];

        const int coordx_idx[] = {i, 1};
        const S kx = coord[coordx_idx];
        const int coordy_idx[] = {i, 0};
        const S ky = coord[coordy_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);

        for (int y = y0; y < y1 + 1; y++) {
            const S wy = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) y - ky) / (width / 2.0));
            for (int x = x0; x < x1 + 1; x++) {
                const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                    fabsf((S) x - kx) / (width / 2.0));
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, mod(y, ny), mod(x, nx)};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, i};
                    output[output_idx] += v;
                }
            }
        }
        """,
        name='interpolate2',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding2_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord', 'raw T output', """
        const int batch_size = output.shape()[0];
        const int ny = output.shape()[1];
        const int nx = output.shape()[2];

        const int coordx_idx[] = {i, 1};
        const S kx = coord[coordx_idx];
        const int coordy_idx[] = {i, 0};
        const S ky = coord[coordy_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);

        for (int y = y0; y < y1 + 1; y++) {
            const S wy = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) y - ky) / (width / 2.0));
            for (int x = x0; x < x1 + 1; x++) {
                const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                    fabsf((S) x - kx) / (width / 2.0));
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, i};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, mod(y, ny), mod(x, nx)};
                    atomicAdd(&output[output_idx], v);
                }
            }
        }
        """, name='gridding2', preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding2_cuda_complex = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = output.shape()[0];
        const int ny = output.shape()[1];
        const int nx = output.shape()[2];

        const int coordx_idx[] = {i, 1};
        const S kx = coord[coordx_idx];
        const int coordy_idx[] = {i, 0};
        const S ky = coord[coordy_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);

        for (int y = y0; y < y1 + 1; y++) {
            const S wy = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) y - ky) / (width / 2.0));
            for (int x = x0; x < x1 + 1; x++) {
                const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                    fabsf((S) x - kx) / (width / 2.0));
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, i};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, mod(y, ny), mod(x, nx)};
                    atomicAdd(reinterpret_cast<T::value_type*>(
                        &(output[output_idx])), v.real());
                    atomicAdd(reinterpret_cast<T::value_type*>(
                        &(output[output_idx])) + 1, v.imag());
                }
            }
        }
        """,
        name='gridding2_complex',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _interpolate3_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord', 'raw T output', """
        const int batch_size = input.shape()[0];
        const int nz = input.shape()[1];
        const int ny = input.shape()[2];
        const int nx = input.shape()[3];

        const int coordz_idx[] = {i, 0};
        const S kz = coord[coordz_idx];
        const int coordy_idx[] = {i, 1};
        const S ky = coord[coordy_idx];
        const int coordx_idx[] = {i, 2};
        const S kx = coord[coordx_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);
        const int z0 = ceil(kz - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);
        const int z1 = floor(kz + width / 2.0);

        for (int z = z0; z < z1 + 1; z++) {
            const S wz = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) z - kz) / (width / 2.0));
            for (int y = y0; y < y1 + 1; y++) {
                const S wy = wz * lin_interpolate(&kernel[0], kernel.size(),
                    fabsf((S) y - ky) / (width / 2.0));
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                        fabsf((S) x - kx) / (width / 2.0));
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, mod(z, nz), mod(y, ny),
                            mod(x, nx)};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, i};
                        output[output_idx] += v;
                    }
                }
            }
        }
        """, name='interpolate3', preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding3_cuda = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord', 'raw T output', """
        const int batch_size = output.shape()[0];
        const int nz = output.shape()[1];
        const int ny = output.shape()[2];
        const int nx = output.shape()[3];

        const int coordz_idx[] = {i, 0};
        const S kz = coord[coordz_idx];
        const int coordy_idx[] = {i, 1};
        const S ky = coord[coordy_idx];
        const int coordx_idx[] = {i, 2};
        const S kx = coord[coordx_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);
        const int z0 = ceil(kz - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);
        const int z1 = floor(kz + width / 2.0);

        for (int z = z0; z < z1 + 1; z++) {
            const S wz = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) z - kz) / (width / 2.0));
            for (int y = y0; y < y1 + 1; y++) {
                const S wy = wz * lin_interpolate(&kernel[0], kernel.size(),
                    fabsf((S) y - ky) / (width / 2.0));
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                        fabsf((S) x - kx) / (width / 2.0));
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, i};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, mod(z, nz), mod(y, ny),
                            mod(x, nx)};
                        atomicAdd(&output[output_idx], v);
                    }
                }
            }
        }
        """, name='gridding3', preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)

    _gridding3_cuda_complex = cp.ElementwiseKernel(
        'raw T input, raw S width, raw S kernel, raw S coord',
        'raw T output',
        """
        const int batch_size = output.shape()[0];
        const int nz = output.shape()[1];
        const int ny = output.shape()[2];
        const int nx = output.shape()[3];

        const int coordz_idx[] = {i, 0};
        const S kz = coord[coordz_idx];
        const int coordy_idx[] = {i, 1};
        const S ky = coord[coordy_idx];
        const int coordx_idx[] = {i, 2};
        const S kx = coord[coordx_idx];

        const int x0 = ceil(kx - width / 2.0);
        const int y0 = ceil(ky - width / 2.0);
        const int z0 = ceil(kz - width / 2.0);

        const int x1 = floor(kx + width / 2.0);
        const int y1 = floor(ky + width / 2.0);
        const int z1 = floor(kz + width / 2.0);

        for (int z = z0; z < z1 + 1; z++) {
            const S wz = lin_interpolate(&kernel[0], kernel.size(),
                fabsf((S) z - kz) / (width / 2.0));
            for (int y = y0; y < y1 + 1; y++) {
                const S wy = wz * lin_interpolate(&kernel[0], kernel.size(),
                     fabsf((S) y - ky) / (width / 2.0));
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * lin_interpolate(&kernel[0], kernel.size(),
                        fabsf((S) x - kx) / (width / 2.0));
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, i};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, mod(z, nz), mod(y, ny),
                            mod(x, nx)};
                        atomicAdd(reinterpret_cast<T::value_type*>(
                            &(output[output_idx])), v.real());
                        atomicAdd(reinterpret_cast<T::value_type*>(
                            &(output[output_idx])) + 1, v.imag());
                    }
                }
            }
        }
        """,
        name='gridding3_complex',
        preamble=lin_interpolate_cuda + mod_cuda,
        reduce_dims=False)



import unittest
import numpy as np
import sigpy as sp
import numpy.testing as npt

from sigpy.mri import app, sim

if __name__ == '__main__':
    unittest.main()


class TestApp(unittest.TestCase):

    def shepp_logan_setup(self):
        img_shape = [6, 6]
        mps_shape = [4, 6, 6]

        img = sp.shepp_logan(img_shape)
        mps = sim.birdcage_maps(mps_shape)

        mask = np.zeros(img_shape)
        mask[:, ::2] = 1

        ksp = mask * sp.fft(mps * img, axes=[-2, -1])
        return img, mps, ksp

    def test_shepp_logan_SenseRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0

        img_rec = app.SenseRecon(
            ksp, mps, lamda, alg_name='ConjugateGradient',
            show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

        img_rec = app.SenseRecon(
            ksp, mps, lamda, alg_name='GradientMethod', show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

        img_rec = app.SenseRecon(
            ksp,
            mps,
            lamda,
            alg_name='PrimalDualHybridGradient',
            max_iter=1000,
            show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    if sp.config.mpi4py_enabled:
        def test_shepp_logan_SenseRecon_with_comm(self):
            img, mps, ksp = self.shepp_logan_setup()
            lamda = 0
            comm = sp.Communicator()
            ksp = ksp[comm.rank::comm.size]
            mps = mps[comm.rank::comm.size]

            img_rec = app.SenseRecon(
                ksp, mps, lamda, comm=comm, alg_name='ConjugateGradient',
                show_pbar=False).run()
            npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

            img_rec = app.SenseRecon(
                ksp, mps, lamda, alg_name='GradientMethod',
                show_pbar=False).run()
            npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

            img_rec = app.SenseRecon(
                ksp,
                mps,
                lamda,
                alg_name='PrimalDualHybridGradient',
                max_iter=1000,
                show_pbar=False).run()
            npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_shepp_logan_SenseConstrainedRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        std = 0

        img_rec = app.SenseConstrainedRecon(ksp, mps, std,
                                            show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_shepp_logan_L1WaveletRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0

        img_rec = app.L1WaveletRecon(
            ksp, mps, lamda, alg_name='GradientMethod',
            show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

        img_rec = app.L1WaveletRecon(
            ksp,
            mps,
            lamda,
            alg_name='PrimalDualHybridGradient',
            max_iter=1000, show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_shepp_logan_L1WaveletConstrainedRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        std = 0

        img_rec = app.L1WaveletConstrainedRecon(
            ksp, mps, std, max_iter=1000, show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_shepp_logan_TotalVariationRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0
        img_rec = app.TotalVariationRecon(ksp, mps, lamda, max_iter=1000,
                                          show_pbar=False).run()

        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_shepp_logan_TotalVariationConstrainedRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        std = 0

        img_rec = app.TotalVariationConstrainedRecon(
            ksp, mps, std, max_iter=2000, show_pbar=False).run()
        npt.assert_allclose(img, img_rec, atol=1e-3, rtol=1e-3)

    def test_ones_JsenseRecon(self):
        img_shape = [6, 6]
        mps_shape = [4, 6, 6]

        img = np.ones(img_shape, dtype=np.complex)
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps * img, axes=[-2, -1])

        _app = app.JsenseRecon(ksp, mps_ker_width=6, ksp_calib_width=6,
                               show_pbar=False)
        mps_rec = _app.run()

        npt.assert_allclose(mps, mps_rec, atol=1e-3, rtol=1e-3)



# -*- coding: utf-8 -*-
"""This module contains an abstract class App for iterative signal processing,
and provides a few general Apps, including a linear least squares App,
and a maximum eigenvalue estimation App.
"""
import numpy as np

from tqdm import tqdm
from sigpy import backend, linop, prox, util
from sigpy.alg import (PowerMethod, GradientMethod,
                       ConjugateGradient, PrimalDualHybridGradient)


class App(object):
    """Abstraction for iterative signal reconstruction applications.

    An App is the final deliverable for each signal reconstruction application.
    The standard way to run an App object, say app, is as follows:

        >>> app.run()

    Each App should have a core Alg object. The run() function runs the Alg,
    with additional convenient features, such as a progress bar, which
    can be toggled with the show_pbar option.

    When creating a new App class, the user should supply an Alg object.
    The user can also optionally define a _pre_update and a _post_update
    function to performs tasks before and after the Alg.update.

    Similar to Alg, an App object is meant to be run once. Different from Alg,
    App is higher level can should use Linop and Prox whenever possible.

    Args:
        alg (Alg): Alg object.
        show_pbar (bool): toggle whether show progress bar.

    Attributes:
        alg (Alg)
        show_pbar (bool)

    """

    def __init__(self, alg, show_pbar=True):
        self.alg = alg
        self.show_pbar = show_pbar

    def _pre_update(self):
        return

    def _post_update(self):
        return

    def _summarize(self):
        return

    def _output(self):
        return

    def run(self):
        if self.show_pbar:
            if self.__class__.__name__ == 'App':
                name = self.alg.__class__.__name__
            else:
                name = self.__class__.__name__

            self.pbar = tqdm(total=self.alg.max_iter, desc=name)

        while(not self.alg.done()):
            self._pre_update()
            self.alg.update()
            self._post_update()
            self._summarize()
            if self.show_pbar:
                self.pbar.update()
                self.pbar.refresh()

        if self.show_pbar:
            self.pbar.close()

        return self._output()


class MaxEig(App):
    """Computes maximum eigenvalue of a Linop.

    Args:
        A (Linop): Hermitian linear operator.
        dtype (Dtype): Data type.
        device (Device): Device.

    Attributes:
        x (int): Eigenvector with largest eigenvalue.

    Output:
        max_eig (int): Largest eigenvalue of A.

    """

    def __init__(self, A, dtype=np.float, device=backend.cpu_device,
                 max_iter=30, show_pbar=True):
        self.x = util.randn(A.ishape, dtype=dtype, device=device)
        alg = PowerMethod(A, self.x, max_iter=max_iter)
        super().__init__(alg, show_pbar=show_pbar)

    def _summarize(self):
        if self.show_pbar:
            self.pbar.set_postfix(max_eig='{0:.2E}'.format(self.alg.max_eig))

    def _output(self):
        return self.alg.max_eig


class LinearLeastSquares(App):
    r"""Linear least squares application.

    Solves for the following problem, with optional regularizations:

    .. math::
        \min_x \frac{1}{2} \| A x - y \|_2^2 + g(G x) +
        \frac{\lambda}{2} \| R x \|_2^2 + \frac{\mu}{2} \| x - z \|_2^2

    Three algorithms can be used: :class:`sigpy.alg.ConjugateGradient`,
    :class:`sigpy.alg.GradientMethod`,
    and :class:`sigpy.alg.PrimalDualHybridGradient`.
    If ``alg_name`` is None, :class:`sigpy.alg.ConjugateGradient` is used
    when ``proxg`` is not specified. If ``proxg`` is specified,
    then :class:`sigpy.alg.GradientMethod` is used when ``G`` is specified,
    and :class:`sigpy.alg.PrimalDualHybridGradient` is used otherwise.

    Args:
        A (Linop): Forward linear operator.
        y (array): Observation.
        x (array): Solution.
        proxg (Prox): Proximal operator of g.
        lamda (float): l2 regularization parameter.
        g (None or function): Regularization function.
            Only used for when `save_objective_values` is true.
        G (None or Linop): Regularization linear operator.
        R (None or Linop): l2 regularization linear operator.
        mu (float): l2 bias regularization parameter.
        z (float or array): Bias for l2 regularization.
        alg_name (str): {`'ConjugateGradient'`, `'GradientMethod'`,
            `'PrimalDualHybridGradient'`}.
        max_iter (int): Maximum number of iterations.
        P (Linop): Preconditioner for ConjugateGradient.
        alpha (None or float): Step size for `GradientMethod`.
        accelerate (bool): Toggle Nesterov acceleration for `GradientMethod`.
        max_power_iter (int): Maximum number of iterations for power method.
            Used for `GradientMethod` when `alpha` is not specified,
            and for `PrimalDualHybridGradient` when `tau` or `sigma` is not
            specified.
        tau (float): Primal step-size for `PrimalDualHybridGradient`.
        sigma (float): Dual step-size for `PrimalDualHybridGradient`.
        save_objective_values (bool): Toggle saving objective value.

    """

    def __init__(self, A, y, x=None, proxg=None,
                 lamda=0, G=None, g=None, R=None, mu=0, z=0,
                 alg_name=None, max_iter=100,
                 P=None, alpha=None, max_power_iter=30, accelerate=True,
                 tau=None, sigma=None,
                 save_objective_values=False, show_pbar=True):
        self.A = A
        self.y = y
        self.x = x
        self.proxg = proxg
        self.lamda = lamda
        self.G = G
        self.g = g
        self.R = R
        self.mu = mu
        self.z = z
        self.alg_name = alg_name
        self.max_iter = max_iter
        self.P = P
        self.alpha = alpha
        self.max_power_iter = max_power_iter
        self.accelerate = accelerate
        self.tau = tau
        self.sigma = sigma
        self.save_objective_values = save_objective_values
        self.show_pbar = show_pbar

        self.y_device = backend.get_device(y)
        if self.x is None:
            with self.y_device:
                self.x = self.y_device.xp.zeros(A.ishape, dtype=y.dtype)

        self.x_device = backend.get_device(self.x)
        self._get_alg()
        if self.save_objective_values:
            self.objective_values = []

    def _summarize(self):
        if self.save_objective_values:
            self.objective_values.append(self.objective())

        if self.show_pbar:
            if self.save_objective_values:
                self.pbar.set_postfix(
                    obj='{0:.2E}'.format(self.objective_values[-1]))
            else:
                self.pbar.set_postfix(resid='{0:.2E}'.format(self.alg.resid))

    def _output(self):
        return self.x

    def _get_alg(self):
        if self.alg_name is None:
            if self.proxg is None:
                self.alg_name = 'ConjugateGradient'
            elif self.G is None:
                self.alg_name = 'GradientMethod'
            else:
                self.alg_name = 'PrimalDualHybridGradient'

        if self.alg_name == 'ConjugateGradient':
            if self.proxg is not None:
                raise ValueError(
                    'ConjugateGradient cannot have proxg specified.')

            self._get_ConjugateGradient()
        elif self.alg_name == 'GradientMethod':
            if self.G is not None:
                raise ValueError('GradientMethod cannot have G specified.')

            self._get_GradientMethod()
        elif self.alg_name == 'PrimalDualHybridGradient':
            self._get_PrimalDualHybridGradient()
        else:
            raise ValueError('Invalid alg_name: {alg_name}.'.format(
                alg_name=self.alg_name))

    def _get_ConjugateGradient(self):
        I = linop.Identity(self.x.shape)
        AHA = self.A.H * self.A
        AHy = self.A.H(self.y)

        if self.lamda != 0:
            if self.R is None:
                AHA += self.lamda * I
            else:
                AHA += self.lamda * self.R.H * self.R

        if self.mu != 0:
            AHA += self.mu * I
            util.axpy(AHy, self.mu, self.z)

        self.alg = ConjugateGradient(
            AHA, AHy, self.x, P=self.P, max_iter=self.max_iter)

    def _get_GradientMethod(self):
        def gradf(x):
            with self.y_device:
                r = self.A(x)
                r -= self.y

            with self.x_device:
                gradf_x = self.A.H(r)
                if self.lamda != 0:
                    if self.R is None:
                        util.axpy(gradf_x, self.lamda, x)
                    else:
                        util.axpy(gradf_x, self.lamda, self.R.H(self.R(x)))

                if self.mu != 0:
                    util.axpy(gradf_x, self.mu, x - self.z)

                return gradf_x

        I = linop.Identity(self.x.shape)
        AHA = self.A.H * self.A

        if self.lamda != 0:
            if self.R is None:
                AHA += self.lamda * I
            else:
                AHA += self.lamda * self.R.H * self.R

        if self.mu != 0:
            AHA += self.mu * I

        max_eig = MaxEig(AHA, dtype=self.x.dtype,
                         device=self.x_device, max_iter=self.max_power_iter,
                         show_pbar=self.show_pbar).run()

        if max_eig == 0:
            self.alpha = 1
        else:
            self.alpha = 1 / max_eig

        self.alg = GradientMethod(
            gradf,
            self.x,
            self.alpha,
            proxg=self.proxg,
            max_iter=self.max_iter,
            accelerate=self.accelerate)

    def _get_PrimalDualHybridGradient(self):
        with self.y_device:
            y = -self.y
            A = self.A

        if self.proxg is None:
            proxg = prox.NoOp(self.x.shape)
        else:
            proxg = self.proxg

        if self.lamda > 0 or self.mu > 0:
            def gradh(x):
                with backend.get_device(self.x):
                    gradh_x = 0
                    if self.lamda > 0:
                        if self.R is None:
                            gradh_x += self.lamda * x
                        else:
                            gradh_x += self.lamda * self.R.H(self.R(x))

                    if self.mu > 0:
                        gradh_x += self.mu * (x - self.z)

                    return gradh_x

            if self.R is None:
                gamma_primal = self.lamda + self.mu
            else:
                gamma_primal = self.mu

        else:
            gradh = None
            gamma_primal = 0

        if self.G is None:
            proxfc = prox.L2Reg(y.shape, 1, y=y)
            gamma_dual = 1
        else:
            A = linop.Vstack([A, self.G])
            proxf1c = prox.L2Reg(self.y.shape, 1, y=y)
            proxf2c = prox.Conj(self.proxg)
            proxfc = prox.Stack([proxf1c, proxf2c])
            proxg = prox.NoOp(self.x.shape)
            gamma_dual = 0

        if self.tau is None:
            if self.sigma is None:
                self.sigma = 1

            S = linop.Multiply(A.oshape, self.sigma)
            AHA = A.H * S * A
            max_eig = MaxEig(
                AHA,
                dtype=self.x.dtype,
                device=self.x_device,
                max_iter=self.max_power_iter,
                show_pbar=self.show_pbar).run()

            self.tau = 1 / (max_eig + self.lamda + self.mu)
        else:
            T = linop.Multiply(A.ishape, self.tau)
            AAH = A * T * A.H

            max_eig = MaxEig(
                AAH,
                dtype=self.x.dtype,
                device=self.x_device,
                max_iter=self.max_power_iter,
                show_pbar=self.show_pbar).run()

            self.sigma = 1 / max_eig

        with self.y_device:
            u = self.y_device.xp.zeros(A.oshape, dtype=self.y.dtype)

        self.alg = PrimalDualHybridGradient(
            proxfc,
            proxg,
            A,
            A.H,
            self.x,
            u,
            self.tau,
            self.sigma,
            gamma_primal=gamma_primal,
            gamma_dual=gamma_dual,
            gradh=gradh,
            max_iter=self.max_iter)

    def objective(self):
        with self.y_device:
            r = self.A(self.x) - self.y

            obj = 1 / 2 * util.norm2(r)
            if self.lamda > 0:
                if self.R is None:
                    obj += self.lamda / 2 * util.norm2(self.x)
                else:
                    obj += self.lamda / 2 * util.norm2(self.R(self.x))

            if self.mu != 0:
                obj += self.mu / 2 * util.norm2(self.x - self.z)

            if self.proxg is not None:
                if self.g is None:
                    raise ValueError(
                        'Cannot compute objective when proxg is specified,'
                        'but g is not.')

                if self.G is None:
                    obj += self.g(self.x)
                else:
                    obj += self.g(self.G(self.x))

            obj = util.asscalar(obj)
            return obj


class L2ConstrainedMinimization(App):
    r"""L2 contrained minimization application.

    Solves for problem:

    .. math::
        &\min_x g(G x) \\
        &\text{s.t.} \| A x - y \|_2 \leq \epsilon

    Args:
        A (Linop): Forward model linear operator.
        y (array): Observation.
        proxg (Prox): Proximal operator of objective.
        eps (float): Residual.

    """

    def __init__(self, A, y, proxg, eps, x=None, G=None,
                 max_iter=100, tau=None, sigma=None,
                 show_pbar=True):
        self.y = y
        self.x = x
        self.y_device = backend.get_device(y)
        if self.x is None:
            with self.y_device:
                self.x = self.y_device.xp.zeros(A.ishape, dtype=self.y.dtype)

        self.x_device = backend.get_device(self.x)
        if G is None:
            self.max_eig_app = MaxEig(
                A.H * A, dtype=self.x.dtype, device=self.x_device,
                show_pbar=show_pbar)

            proxfc = prox.Conj(prox.L2Proj(A.oshape, eps, y=y))
        else:
            proxf1 = prox.L2Proj(A.oshape, eps, y=y)
            proxf2 = proxg
            proxfc = prox.Conj(prox.Stack([proxf1, proxf2]))
            proxg = prox.NoOp(A.ishape)
            A = linop.Vstack([A, G])

        if tau is None or sigma is None:
            max_eig = MaxEig(A.H * A, dtype=self.x.dtype,
                             device=self.x_device,
                             show_pbar=show_pbar).run()
            tau = 1
            sigma = 1 / max_eig

        with self.y_device:
            self.u = self.y_device.xp.zeros(A.oshape, dtype=self.y.dtype)

        alg = PrimalDualHybridGradient(proxfc, proxg, A, A.H, self.x, self.u,
                                       tau, sigma, max_iter=max_iter)

        super().__init__(alg, show_pbar=show_pbar)

    def _summarize(self):
        if self.show_pbar:
            self.pbar.set_postfix(resid='{0:.2E}'.format(self.alg.resid))

    def _output(self):
        return self.x



"""The module contains functions and classes for building iterative
reconstruction applications for MRI.

It provides convenient simulation and sampling functions,
such as poisson-disc sampling function, and shepp-logan
phantom generation function. It also implements common
MRI reconstruction applications, including SENSE reconstruction,
l1-wavelet reconstruction, total-variation reconstruction,
and JSENSE reconstruction.

"""
from sigpy.mri import app, linop

from sigpy.mri import espirit, precond, samp, sim, util
from sigpy.mri.espirit import *  # noqa
from sigpy.mri.precond import *  # noqa
from sigpy.mri.samp import *  # noqa
from sigpy.mri.sim import *  # noqa
from sigpy.mri.util import *  # noqa

__all__ = ['app', 'linop']
__all__.extend(espirit.__all__)
__all__.extend(precond.__all__)
__all__.extend(samp.__all__)
__all__.extend(sim.__all__)
__all__.extend(util.__all__)



import unittest
import numpy as np
from sigpy import interp, config

if config.cupy_enabled:
    import cupy as cp

if __name__ == '__main__':
    unittest.main()


class TestInterp(unittest.TestCase):

    def test_interpolate(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        batch = 2
        for xp in xps:
            for ndim in [1, 2, 3]:
                with self.subTest(ndim=ndim, xp=xp):
                    shape = [3] + [1] * (ndim - 1)
                    width = 2.0
                    kernel = xp.array([1.0, 0.5])
                    coord = xp.array([[0.1] + [0] * (ndim - 1),
                                      [1.1] + [0] * (ndim - 1),
                                      [2.1] + [0] * (ndim - 1)])

                    input = xp.array([[0, 1.0, 0]] * batch).reshape(
                        [batch] + shape)
                    output = interp.interpolate(input, width, kernel, coord)
                    output_expected = xp.array([[0.1, 0.9, 0]] * batch)
                    xp.testing.assert_allclose(output, output_expected,
                                               atol=1e-7)

    def test_gridding(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        batch = 2
        for xp in xps:
            for ndim in [1, 2, 3]:
                with self.subTest(ndim=ndim, xp=xp):
                    shape = [3] + [1] * (ndim - 1)
                    width = 2.0
                    kernel = xp.array([1.0, 0.5])
                    coord = xp.array([[0.1] + [0] * (ndim - 1),
                                      [1.1] + [0] * (ndim - 1),
                                      [2.1] + [0] * (ndim - 1)])

                    input = xp.array([[0, 1.0, 0]] * batch)
                    output = interp.gridding(
                        input, [batch] + shape, width, kernel, coord)
                    output_expected = xp.array(
                        [[0, 0.9, 0.1]] * batch).reshape([batch] + shape)
                    xp.testing.assert_allclose(output, output_expected,
                                               atol=1e-7)



# -*- coding: utf-8 -*-
"""MRI sampling functions.
"""
import numpy as np
import numba as nb


__all__ = ['poisson', 'radial']


def poisson(img_shape, accel, K=30, calib=[0, 0], dtype=np.complex,
            crop_corner=True, return_density=False, seed=0):
    """Generate Poisson-disc sampling pattern

    Args:
        img_shape (tuple of ints): length-2 image shape.
        accel (float): Target acceleration factor. Greater than 1.
        K (float): maximum number of samples to reject.
        calib (tuple of ints): length-2 calibration shape.
        dtype (Dtype): data type.
        crop_corner (bool): Toggle whether to crop sampling corners.
        return_density (bool): Toggle whether to return sampling density.
        seed (int): Random seed.

    Returns:
        array: Poisson-disc sampling mask.

    References:
        Bridson, Robert. "Fast Poisson disk sampling in arbitrary dimensions."
        SIGGRAPH sketches. 2007.

    """
    y, x = np.mgrid[:img_shape[-2], :img_shape[-1]]
    x = np.maximum(abs(x - img_shape[-1] / 2) - calib[-1] / 2, 0)
    x /= x.max()
    y = np.maximum(abs(y - img_shape[-2] / 2) - calib[-2] / 2, 0)
    y /= y.max()
    r = np.sqrt(x ** 2 + y ** 2)

    slope_max = 40
    slope_min = 0
    while slope_min < slope_max:
        slope = (slope_max + slope_min) / 2.0
        R = (1.0 + r * slope)
        mask = _poisson(img_shape[-1], img_shape[-2], K, R, calib, seed)
        if crop_corner:
            mask *= r < 1

        est_accel = img_shape[-1] * img_shape[-2] / np.sum(mask[:])

        if abs(est_accel - accel) < 0.1:
            break
        if est_accel < accel:
            slope_min = slope
        else:
            slope_max = slope

    mask = mask.reshape(img_shape).astype(dtype)
    if return_density:
        return mask, r
    else:
        return mask


def radial(coord_shape, img_shape, golden=True, dtype=np.float):
    """Generate radial trajectory.

    Args:
        coord_shape (tuple of ints): coordinates of shape [ntr, nro, ndim]
        img_shape (tuple of ints): image shape.
        golden (bool): golden angle ordering.
        dtype (Dtype): data type.

    Returns:
        array: radial coordinates.

    """
    ntr, nro, ndim = coord_shape

    if ndim == 2:
        if golden:
            # 111.25 degrees in radians
            # from: Winkelmann, S. An Optimal Radial Profile Order Based on
            # the Golden Ratio for Time-Resolved MRI, IEEE TMI, 2007
            phi = 0.970839395
        else:
            phi = 1.0 / ntr

        n, r = np.mgrid[:ntr, :0.5:0.5 / nro]

        theta = 2.0 * np.pi * n * phi
        coord = np.zeros((ntr, nro, 2))
        coord[:, :, -1] = r * np.cos(theta)
        coord[:, :, -2] = r * np.sin(theta)

    elif ndim == 3:
        if golden:
            phi1 = 0.465571231876768
            phi2 = 0.682327803828019
        else:
            raise NotImplementedError

        n, r = np.mgrid[:ntr, :0.5:0.5 / nro]

        beta = np.arccos(2 * ((n * phi1) % 1) - 1)
        alpha = 2 * np.pi * ((n * phi2) % 1)

        coord = np.zeros((ntr, nro, 3))
        coord[:, :, -1] = r * np.sin(beta) * np.cos(alpha)
        coord[:, :, -2] = r * np.sin(beta) * np.sin(alpha)
        coord[:, :, -3] = r * np.cos(beta)
    else:
        raise ValueError('Invalid ndim: {}'.format(ndim))

    return (coord * img_shape[-ndim:]).astype(dtype)


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _poisson(nx, ny, K, R, calib, seed):

    mask = np.zeros((ny, nx))
    f = ny / nx

    if seed is not None:
        np.random.seed(int(seed))

    pxs = np.empty(nx * ny, np.int32)
    pys = np.empty(nx * ny, np.int32)
    pxs[0] = np.random.randint(0, nx)
    pys[0] = np.random.randint(0, ny)
    m = 1
    while (m > 0):

        i = np.random.randint(0, m)
        px = pxs[i]
        py = pys[i]
        rad = R[py, px]

        # Attempt to generate point
        done = False
        k = 0
        while not done and k < K:

            # Generate point randomly from R and 2R
            rd = rad * (np.random.random() * 3 + 1)**0.5
            t = 2 * np.pi * np.random.random()
            qx = px + rd * np.cos(t)
            qy = py + rd * f * np.sin(t)

            # Reject if outside grid or close to other points
            if qx >= 0 and qx < nx and qy >= 0 and qy < ny:

                startx = max(int(qx - rad), 0)
                endx = min(int(qx + rad + 1), nx)
                starty = max(int(qy - rad * f), 0)
                endy = min(int(qy + rad * f + 1), ny)

                done = True
                for x in range(startx, endx):
                    for y in range(starty, endy):
                        if (mask[y, x] == 1
                            and (((qx - x) / R[y, x]) ** 2 +
                                 ((qy - y) / (R[y, x] * f)) ** 2 < 1)):
                            done = False
                            break

                    if not done:
                        break

            k += 1

        # Add point if done else remove active
        if done:
            pxs[m] = qx
            pys[m] = qy
            mask[int(qy), int(qx)] = 1
            m += 1
        else:
            pxs[i] = pxs[m - 1]
            pys[i] = pys[m - 1]
            m -= 1

    # Add calibration region
    mask[int(ny / 2 - calib[-2] / 2):int(ny / 2 + calib[-2] / 2),
         int(nx / 2 - calib[-1] / 2):int(nx / 2 + calib[-1] / 2)] = 1

    return mask



# -*- coding: utf-8 -*-
"""Thresholding functions.
"""
import numpy as np
import numba as nb

from sigpy import backend, config, util


__all__ = ['soft_thresh', 'hard_thresh', 'l1_proj', 'l2_proj']


def soft_thresh(lamda, input):
    r"""Soft threshold.

    Performs:

    .. math::
        (| x | - \lambda)_+  \text{sgn}(x)

    Args:
        lamda (float, or array): Threshold parameter.
        input (array)

    Returns:
        array: soft-thresholded result.

    """
    device = backend.get_device(input)
    xp = device.xp

    lamda = xp.real(lamda)
    with device:
        if device == backend.cpu_device:
            output = _soft_thresh(lamda, input)
        else:  # pragma: no cover
            output = _soft_thresh_cuda(lamda, input)

        if np.issubdtype(input.dtype, np.floating):
            output = xp.real(output)

    return output


def hard_thresh(lamda, input):
    """Hard threshold.

    Args:
        lamda (float, or array): Threshold parameter.
        input (array)

    Returns:
        array: hard-thresholded result.

    """
    device = backend.get_device(input)

    if device == backend.cpu_device:
        return _hard_thresh(lamda, input)
    else:  # pragma: no cover
        with device:
            return _hard_thresh_cuda(lamda, input)


def l1_proj(eps, input):
    """Projection onto L1 ball.

    Args:
        eps (float, or array): L1 ball scaling.
        input (array)

    Returns:
        array: Result.

    References:
        J. Duchi, S. Shalev-Shwartz, and Y. Singer, "Efficient projections onto
        the l1-ball for learning in high dimensions" 2008.

    """
    device = backend.get_device(input)
    xp = device.xp

    with device:
        shape = input.shape
        input = input.ravel()

        if xp.linalg.norm(input, 1) < eps:
            return input
        else:
            shape = len(input)
            s = xp.sort(xp.abs(input))[::-1]
            st = (xp.cumsum(s) - eps) / (xp.arange(shape) + 1)
            idx = xp.flatnonzero((s - st) > 0).max()
            return soft_thresh(st[idx], input.reshape(shape))


def l2_proj(eps, input, axes=None):
    """Projection onto L2 ball.

    Args:
        eps (float, or array): L2 ball scaling.
        input (array)

    Returns:
        array: Result.

    """
    axes = util._normalize_axes(axes, input.ndim)

    device = backend.get_device(input)
    xp = device.xp
    with device:
        norm = xp.sum(xp.abs(input)**2, axis=axes, keepdims=True)**0.5
        mask = norm < eps
        output = mask * input + (1 - mask) * (eps * input / (norm + mask))

    return output


@nb.vectorize  # pragma: no cover
def _soft_thresh(lamda, input):
    abs_input = abs(input)
    if (abs_input == 0):
        sign = 0
    else:
        sign = input / abs_input

    mag = abs_input - lamda
    mag = (abs(mag) + mag) / 2

    return mag * sign


@nb.vectorize  # pragma: no cover
def _hard_thresh(lamda, input):
    abs_input = abs(input)
    if abs_input > lamda:
        return input
    else:
        return 0


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _soft_thresh_cuda = cp.ElementwiseKernel(
        'S lamda, T input',
        'T output',
        """
        S abs_input = abs(input);
        T sign;
        if (abs_input == 0)
            sign = 0;
        else
            sign = input / (T) abs_input;
        S mag = abs_input - lamda;
        mag = (abs(mag) + mag) / 2.;

        output = (T) mag * sign;
        """,
        name='soft_thresh')

    _hard_thresh_cuda = cp.ElementwiseKernel(
        'S lamda, T input',
        'T output',
        """
        S abs_input = abs(input);
        if (abs_input > lamda)
            output = input;
        else
            output = 0;
        """,
        name='hard_thresh')



import sigpy as sp


__all__ = ['espirit_maps']


def espirit_maps(ksp, calib_width=24,
                 thresh=0.001, kernel_width=6,
                 crop=0.8,
                 max_power_iter=30, device=sp.cpu_device,
                 output_eigenvalue=False):
    """Generate ESPIRiT maps from k-space.

    Currently only supports outputting one set of maps.

    Args:
        ksp (array): k-space array of shape [num_coils, n_ndim, ..., n_1]
        calib (tuple of ints): length-2 image shape.
        thresh (float): threshold for the calibration matrix.
        kernel_width (int): kernel width for the calibration matrix.
        max_power_iter (int): maximum number of power iterations.
        device (Device): computing device.
        crop (int): cropping threshold.

    Returns:
        array: ESPIRiT maps of the same shape as ksp.

    References:
        Martin Uecker, Peng Lai, Mark J. Murphy, Patrick Virtue, Michael Elad,
        John M. Pauly, Shreyas S. Vasanawala, and Michael Lustig
        ESPIRIT - An Eigenvalue Approach to Autocalibrating Parallel MRI:
        Where SENSE meets GRAPPA.
        Magnetic Resonance in Medicine, 71:990-1001 (2014)

    """
    img_ndim = ksp.ndim - 1
    num_coils = len(ksp)
    with sp.get_device(ksp):
        # Get calibration region
        calib_shape = [num_coils] + [calib_width] * img_ndim
        calib = sp.resize(ksp, calib_shape)
        calib = sp.to_device(calib, device)

    device = sp.Device(device)
    xp = device.xp
    with device:
        # Get calibration matrix
        kernel_shape = [num_coils] + [kernel_width] * img_ndim
        kernel_strides = [1] * (img_ndim + 1)
        mat = sp.array_to_blocks(calib, kernel_shape, kernel_strides)
        mat = mat.reshape([-1, sp.prod(kernel_shape)])

        # Perform SVD on calibration matrix
        _, S, VH = xp.linalg.svd(mat, full_matrices=False)
        VH = VH[S > thresh * S.max(), :]

        # Get kernels
        num_kernels = len(VH)
        kernels = VH.reshape([num_kernels] + kernel_shape)
        img_shape = ksp.shape[1:]

        # Get covariance matrix in image domain
        AHA = xp.zeros(img_shape[::-1] + (num_coils, num_coils),
                       dtype=ksp.dtype)
        for kernel in kernels:
            img_kernel = sp.ifft(sp.resize(kernel, ksp.shape),
                                 axes=range(-img_ndim, 0))
            aH = xp.expand_dims(img_kernel.T, axis=-1)
            a = xp.conj(aH.swapaxes(-1, -2))
            AHA += aH @ a

        AHA *= (sp.prod(img_shape) / kernel_width**img_ndim)

        # Power Iteration to compute top eigenvector
        mps = xp.ones(ksp.shape[::-1] + (1, ), dtype=ksp.dtype)
        for _ in range(max_power_iter):
            sp.copyto(mps, AHA @ mps)
            eig_value = xp.sum(xp.abs(mps)**2, axis=-2, keepdims=True)**0.5
            mps /= eig_value

        # Normalize phase with respect to first channel
        mps = mps.T[0]
        mps *= xp.conj(mps[0] / xp.abs(mps[0]))

        # Crop maps by thresholding eigenvalue
        eig_value = eig_value.T[0]
        mps *= eig_value > crop

        if output_eigenvalue:
            return mps, eig_value
        else:
            return mps
